# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import functools
import inspect
import os
import pathlib
import shutil
from datetime import datetime
from importlib.metadata import version

import fsspec
import gcsfs
import nest_asyncio
import numpy as np
import xarray as xr
import zarr
from fsspec.implementations.cached import WholeFileCacheFileSystem
from loguru import logger
from tqdm.asyncio import tqdm

from earth2studio.data.utils import (
    AsyncCachingFileSystem,
    datasource_cache_root,
    prep_data_inputs,
)
from earth2studio.lexicon import ARCOLexicon
from earth2studio.utils.type import TimeArray, VariableArray


class ARCO:
    """Analysis-Ready, Cloud Optimized (ARCO) is a data store of ERA5 re-analysis data
    currated by Google. This data is stored in Zarr format and contains 31 surface and
    pressure level variables (for 37 pressure levels)  on a 0.25 degree lat lon grid.
    Temporal resolution is 1 hour.

    Parameters
    ----------
    cache : bool, optional
        Cache data source on local memory, by default True
    verbose : bool, optional
        Print download progress, by default True
    async_timeout : int, optional
        Time in sec after which download will be cancelled if not finished successfully,
        by default 600

    Warning
    -------
    This is a remote data source and can potentially download a large amount of data
    to your local machine for large requests.

    Note
    ----
    Additional information on the data repository can be referenced here:

    - https://cloud.google.com/storage/docs/public-datasets/era5
    """

    ARCO_LAT = np.linspace(90, -90, 721)
    ARCO_LON = np.linspace(0, 359.75, 1440)

    def __init__(
        self, cache: bool = True, verbose: bool = True, async_timeout: int = 600
    ):
        # Check Zarr version and use appropriate method
        try:
            zarr_version = version("zarr")
            self.zarr_major_version = int(zarr_version.split(".")[0])
        except Exception:
            # Fallback to older method if version check fails
            self.zarr_major_version = 2  # Assume older version if we can't determine

        self._cache = cache
        self._verbose = verbose

        fs = gcsfs.GCSFileSystem(
            cache_timeout=-1,
            token="anon",  # noqa: S106 # nosec B106
            access="read_only",
            block_size=8**20,
            asynchronous=(self.zarr_major_version == 3),
        )

        if self._cache:
            cache_options = {
                "cache_storage": self.cache,
                "expiry_time": 31622400,  # 1 year
            }
            if self.zarr_major_version == 3:
                fs = AsyncCachingFileSystem(fs=fs, **cache_options, asynchronous=True)
            else:
                fs = WholeFileCacheFileSystem(fs=fs, **cache_options)

        if self.zarr_major_version >= 3:
            # Zarr 3.0+ method
            zstore = zarr.storage.FsspecStore(
                fs,
                path="/gcp-public-data-arco-era5/ar/full_37-1h-0p25deg-chunk-1.zarr-v3",
            )
            self.zarr_group = zarr.api.asynchronous.open(store=zstore, mode="r")
            self.level_coords = None
        else:
            # Legacy method for Zarr < 3.0
            logger.warning(
                "Using Zarr 2.0 method for ARCO, this can be extremely slow with caching!"
            )
            fs_map = fsspec.FSMap(
                "gcp-public-data-arco-era5/ar/full_37-1h-0p25deg-chunk-1.zarr-v3", fs
            )
            self.zarr_group = zarr.open(fs_map, mode="r")
            self.level_coords = self.zarr_group["level"][:]

        self.async_timeout = async_timeout

    def __call__(
        self,
        time: datetime | list[datetime] | TimeArray,
        variable: str | list[str] | VariableArray,
    ) -> xr.DataArray:
        """Function to get data

        Parameters
        ----------
        time : datetime | list[datetime] | TimeArray
            Timestamps to return data for (UTC).
        variable : str | list[str] | VariableArray
            String, list of strings or array of strings that refer to variables to
            return. Must be in the ARCO lexicon.

        Returns
        -------
        xr.DataArray
            ERA5 weather data array from ARCO
        """

        nest_asyncio.apply()  # Patch asyncio to work in notebooks
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            # If no event loop exists, create one
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        xr_array = loop.run_until_complete(
            asyncio.wait_for(self.fetch(time, variable), timeout=self.async_timeout)
        )

        # Delete cache if needed
        if not self._cache:
            shutil.rmtree(self.cache)

        return xr_array

    async def fetch(
        self,
        time: datetime | list[datetime] | TimeArray,
        variable: str | list[str] | VariableArray,
    ) -> xr.DataArray:
        """Async function to get data

        Parameters
        ----------
        time : datetime | list[datetime] | TimeArray
            Timestamps to return data for (UTC).
        variable : str | list[str] | VariableArray
            String, list of strings or array of strings that refer to variables to
            return. Must be in the ARCO lexicon.

        Returns
        -------
        xr.DataArray
            ERA5 weather data array from ARCO
        """
        time, variable = prep_data_inputs(time, variable)
        # Create cache dir if doesnt exist
        pathlib.Path(self.cache).mkdir(parents=True, exist_ok=True)

        # Make sure input time is valid
        self._validate_time(time)

        xr_array = xr.DataArray(
            data=np.empty(
                (len(time), len(variable), len(self.ARCO_LAT), len(self.ARCO_LON))
            ),
            dims=["time", "variable", "lat", "lon"],
            coords={
                "time": time,
                "variable": variable,
                "lat": self.ARCO_LAT,
                "lon": self.ARCO_LON,
            },
        )

        args = [
            (t, i, v, j) for j, v in enumerate(variable) for i, t in enumerate(time)
        ]
        func_map = map(functools.partial(self.fetch_wrapper, xr_array=xr_array), args)

        # Before anything wait until the group gets opened
        if self.zarr_major_version >= 3:
            if inspect.isawaitable(self.zarr_group):
                self.zarr_group = await self.zarr_group
            self.level_coords = await (await self.zarr_group.get("level")).getitem(
                slice(None)
            )
        # Launch all fetch requests
        await tqdm.gather(
            *func_map, desc="Fetching ARCO data", disable=(not self._verbose)
        )
        return xr_array

    async def fetch_wrapper(
        self,
        e: tuple[datetime, int, str, int],
        xr_array: xr.DataArray,
    ) -> None:
        """Small wrapper to pack arrays into the DataArray"""
        out = await self.fetch_array(e[0], e[2])
        xr_array[e[1], e[3]] = out

    async def fetch_array(self, time: datetime, variable: str) -> np.ndarray:
        """Fetches requested array from remote store

        Parameters
        ----------
        time : datetime
            Time to fetch
        variable : str
            Variable to fetch

        Returns
        -------
        np.ndarray
            Data
        """
        # Get time index (vanilla zarr doesnt support date indices)
        time_index = self._get_time_index(time)
        logger.debug(
            f"Fetching ARCO zarr array for variable: {variable} at {time.isoformat()}"
        )
        try:
            arco_name, modifier = ARCOLexicon[variable]
        except KeyError as e:
            logger.error(f"variable id {variable} not found in ARCO lexicon")
            raise e

        arco_variable, level = arco_name.split("::")
        if self.zarr_major_version >= 3:
            zarr_array = await self.zarr_group.get(arco_variable)
            shape = zarr_array.shape
            # Static variables
            if len(shape) == 2:
                data = await zarr_array.getitem(slice(None))
                output = modifier(data)
            # Surface variable
            elif len(shape) == 3:
                data = await zarr_array.getitem(time_index)
                output = modifier(data)
            # Atmospheric variable
            else:
                # Load levels coordinate system from Zarr store and check
                level_index = np.searchsorted(self.level_coords, int(level))
                data = await zarr_array.getitem((time_index, level_index))
                output = modifier(data)
        else:
            # Zarr 2.0 fall back
            shape = self.zarr_group[arco_variable].shape
            # Static variables
            if len(shape) == 2:
                output = modifier(self.zarr_group[arco_variable][:])
            # Surface variable
            elif len(shape) == 3:
                output = modifier(self.zarr_group[arco_variable][time_index])
            # Atmospheric variable
            else:
                level_index = np.where(self.level_coords == int(level))[0][0]
                output = modifier(
                    self.zarr_group[arco_variable][time_index, level_index]
                )

        return output

    @property
    def cache(self) -> str:
        """Get the appropriate cache location."""
        cache_location = os.path.join(datasource_cache_root(), "arco")
        if not self._cache:
            cache_location = os.path.join(cache_location, "tmp_arco")
        return cache_location

    @classmethod
    def _validate_time(cls, times: list[datetime]) -> None:
        """Verify if date time is valid for ARCO

        Parameters
        ----------
        times : list[datetime]
            list of date times to fetch data
        """
        for time in times:
            if not (time - datetime(1900, 1, 1)).total_seconds() % 3600 == 0:
                raise ValueError(
                    f"Requested date time {time} needs to be 1 hour interval for ARCO"
                )

            if time < datetime(year=1940, month=1, day=1):
                raise ValueError(
                    f"Requested date time {time} needs to be after January 1st, 1940 for ARCO"
                )

            if time >= datetime(year=2023, month=11, day=10):
                raise ValueError(
                    f"Requested date time {time} needs to be before November 10th, 2023 for ARCO"
                )

            # if not self.available(time):
            #     raise ValueError(f"Requested date time {time} not available in ARCO")

    @classmethod
    def _get_time_index(cls, time: datetime) -> int:
        """Small little index converter to go from datetime to integer index.
        We don't need to do with with xarray, but since we are vanilla zarr for speed
        this conversion must be manual.

        Parameters
        ----------
        time : datetime
            Input date time

        Returns
        -------
        int
            Time coordinate index of ARCO data
        """
        start_date = datetime(year=1900, month=1, day=1)
        duration = time - start_date
        return int(divmod(duration.total_seconds(), 3600)[0])

    @classmethod
    def available(cls, time: datetime | np.datetime64) -> bool:
        """Checks if given date time is avaliable in the ARCO data source

        Parameters
        ----------
        time : datetime | np.datetime64
            Date time to access

        Returns
        -------
        bool
            If date time is avaiable
        """
        if isinstance(time, np.datetime64):  # np.datetime64 -> datetime
            _unix = np.datetime64(0, "s")
            _ds = np.timedelta64(1, "s")
            time = datetime.utcfromtimestamp(float((time - _unix) / _ds))

        # Offline checks
        try:
            cls._validate_time([time])
        except ValueError:
            return False

        gcs = gcsfs.GCSFileSystem(cache_timeout=-1)

        try:
            zarr_version = version("zarr")
            zarr_major_version = int(zarr_version.split(".")[0])
        except Exception:
            zarr_major_version = 2

        if zarr_major_version >= 3:
            gcstore = zarr.storage.FsspecStore(
                gcs,
                path="/gcp-public-data-arco-era5/ar/full_37-1h-0p25deg-chunk-1.zarr-v3",
            )
        else:
            gcstore = gcsfs.GCSMap(
                "gs://gcp-public-data-arco-era5/ar/full_37-1h-0p25deg-chunk-1.zarr-v3",
                gcs=gcs,
            )

        zarr_group = zarr.open(gcstore, mode="r")
        # Load time coordinate system from Zarr store and check
        time_index = cls._get_time_index(time)
        max_index = zarr_group["time"][-1]
        return time_index >= 0 and time_index <= max_index



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import concurrent.futures
import functools
import hashlib
import os
import pathlib
import shutil
from collections.abc import Callable
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone

import nest_asyncio
import numpy as np
import s3fs
import xarray as xr
from loguru import logger
from tqdm.asyncio import tqdm

from earth2studio.data.utils import (
    datasource_cache_root,
    prep_forecast_inputs,
)
from earth2studio.lexicon import GEFSLexicon, GEFSLexiconSel
from earth2studio.utils.type import LeadTimeArray, TimeArray, VariableArray

logger.remove()
logger.add(lambda msg: tqdm.write(msg, end=""), colorize=True)


@dataclass
class GEFSAsyncTask:
    """Small helper struct for Async tasks"""

    data_array_indices: tuple[int, int, int]
    gefs_file_uri: str
    gefs_byte_offset: int
    gefs_byte_length: int | None
    gefs_modifier: Callable


class GEFS_FX:
    """The Global Ensemble Forecast System (GEFS) forecast source is a 30 member
    ensemble forecast provided on an 0.5 degree equirectangular grid.  GEFS is a weather
    forecast model developed by  National Centers for Environmental Prediction (NCEP).
    This forecast source has data at 6-hour intervals spanning from Sept 23rd 2020 to
    present date. Each forecast provides 3-hourly predictions up to 10 days (240 hours)
    and 6 hourly predictions for another 6 days (384 hours).

    Parameters
    ----------
    member : str, optional
        GEFS member. Options are: control gec00 (control), gepNN (forecast member NN,
        e.g. gep01, gep02,...), by default "gec00"
    max_workers : int, optional
        Max works in async io thread pool. Only applied when using sync call function
        and will modify the default async loop if one exists, by default 24
    cache : bool, optional
        Cache data source on local memory, by default True
    verbose : bool, optional
        Print download progress, by default True
    async_timeout : int, optional
        Time in sec after which download will be cancelled if not finished successfully,
        by default 600

    Warning
    -------
    This is a remote data source and can potentially download a large amount of data
    to your local machine for large requests.

    Warning
    -------
    Some variables in the GEFS lexicon may not be available at lead time 0. Consult GEFS
    documentation for additional information.

    Note
    ----
    Additional information on the data repository can be referenced here:

    - https://registry.opendata.aws/noaa-gefs/
    - https://www.ncei.noaa.gov/products/weather-climate-models/global-ensemble-forecast
    - https://www.nco.ncep.noaa.gov/pmb/products/gens/
    """

    GEFS_BUCKET_NAME = "noaa-gefs-pds"
    MAX_BYTE_SIZE = 5000000

    GEFS_LAT = np.linspace(90, -90, 361)
    GEFS_LON = np.linspace(0, 359.5, 720)

    GEFS_MEMBERS = ["gec00"] + [f"gep{i:02d}" for i in range(1, 31)]
    GEFS_CHECK_PRODUCT = "pgrb2bp5"

    def __init__(
        self,
        member: str = "gec00",
        max_workers: int = 24,
        cache: bool = True,
        verbose: bool = True,
        async_timeout: int = 600,
    ):
        self._cache = cache
        self._verbose = verbose
        self._max_workers = max_workers

        if member not in self.GEFS_MEMBERS:
            raise ValueError(f"Invalid GEFS member {member}")

        self._product_class = "atmos"
        self._product_resolution = "0p50"  # 0p50 or 0p25
        self._member = member
        self.fs = s3fs.S3FileSystem(anon=True, client_kwargs={}, asynchronous=True)
        self.async_timeout = async_timeout
        self.lexicon = GEFSLexicon

    def __call__(
        self,
        time: datetime | list[datetime] | TimeArray,
        lead_time: timedelta | list[timedelta] | LeadTimeArray,
        variable: str | list[str] | VariableArray,
    ) -> xr.DataArray:
        """Retrieve GEFS forecast data

        Parameters
        ----------
        time : datetime | list[datetime] | TimeArray
            Timestamps to return data for (UTC).
        lead_time: timedelta | list[timedelta] | LeadTimeArray
            Forecast lead times to fetch.
        variable : str | list[str] | VariableArray
            String, list of strings or array of strings that refer to variables to
            return. Must be in the GEFS lexicon.

        Returns
        -------
        xr.DataArray
            GEFS forecast data array
        """
        nest_asyncio.apply()  # Patch asyncio to work in notebooks
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            # If no event loop exists, create one
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        # Modify the worker amount
        loop.set_default_executor(
            concurrent.futures.ThreadPoolExecutor(max_workers=self._max_workers)
        )

        xr_array = loop.run_until_complete(
            asyncio.wait_for(
                self.fetch(time, lead_time, variable), timeout=self.async_timeout
            )
        )

        return xr_array

    async def fetch(
        self,
        time: datetime | list[datetime] | TimeArray,
        lead_time: timedelta | list[timedelta] | LeadTimeArray,
        variable: str | list[str] | VariableArray,
    ) -> xr.DataArray:
        """Async function to get data

        Parameters
        ----------
        time : datetime | list[datetime] | TimeArray
            Timestamps to return data for (UTC).
        lead_time: timedelta | list[timedelta] | LeadTimeArray
            Forecast lead times to fetch.
        variable : str | list[str] | VariableArray
            String, list of strings or array of strings that refer to variables to
            return. Must be in the GEFS FX lexicon.

        Returns
        -------
        xr.DataArray
            GEFS forecast data array
        """
        time, lead_time, variable = prep_forecast_inputs(time, lead_time, variable)
        # Create cache dir if doesnt exist
        pathlib.Path(self.cache).mkdir(parents=True, exist_ok=True)

        # Make sure input time is valid
        self._validate_time(time)
        self._validate_leadtime(lead_time)

        # Note, this could be more memory efficient and avoid pre-allocation of the array
        # but this is much much cleaner to deal with, compared to something seen in the
        # NCAR data source.
        xr_array = xr.DataArray(
            data=np.empty(
                (
                    len(time),
                    len(lead_time),
                    len(variable),
                    len(self.GEFS_LAT),
                    len(self.GEFS_LON),
                )
            ),
            dims=["time", "lead_time", "variable", "lat", "lon"],
            coords={
                "time": time,
                "lead_time": lead_time,
                "variable": variable,
                "lat": self.GEFS_LAT,
                "lon": self.GEFS_LON,
            },
        )

        async_tasks = []
        async_tasks = await self._create_tasks(time, lead_time, variable)
        func_map = map(
            functools.partial(self.fetch_wrapper, xr_array=xr_array), async_tasks
        )

        await tqdm.gather(
            *func_map, desc="Fetching GEFS data", disable=(not self._verbose)
        )

        # Delete cache if needed
        if not self._cache:
            shutil.rmtree(self.cache)

        # Close aiohttp client if s3fs
        # https://github.com/fsspec/s3fs/issues/943
        # https://github.com/zarr-developers/zarr-python/issues/2901
        await self.fs.set_session()  # Make sure the session was actually initalized
        s3fs.S3FileSystem.close_session(asyncio.get_event_loop(), self.fs.s3)

        return xr_array

    async def _create_tasks(
        self, time: list[datetime], lead_time: list[timedelta], variable: list[str]
    ) -> list[GEFSAsyncTask]:
        """Create download tasks, each corresponding to one grib byte range

        Parameters
        ----------
        times : list[datetime]
            Timestamps to be downloaded (UTC).
        variables : list[str]
            List of variables to be downloaded.

        Returns
        -------
        list[dict]
            List of download tasks
        """
        tasks: list[GEFSAsyncTask] = []  # group pressure-level variables

        # Start with fetching all index files for each time / lead time
        products = set()
        for v in variable:
            gefs_name_str, _ = self.lexicon[v]  # type: ignore
            products.add(gefs_name_str.split("::")[0])
        args = [
            self._grib_index_uri(t, lt, p)
            for t in time
            for lt in lead_time
            for p in products
        ]
        func_map = map(self._fetch_index, args)
        results = await tqdm.gather(
            *func_map, desc="Fetching GEFS index files", disable=True
        )
        for i, t in enumerate(time):
            for j, lt in enumerate(lead_time):
                # Get index file dictionaries for each possible product
                index_files = {p: results.pop(0) for p in products}
                for k, v in enumerate(variable):
                    try:
                        gefs_name_str, modifier = self.lexicon[v]  # type: ignore
                        gefs_name = gefs_name_str.split("::") + ["", ""]
                        product = gefs_name[0]
                        variable_name = gefs_name[1]
                        level = gefs_name[2]

                        # Create index key to find byte range
                        gefs_key = f"{variable_name}::{level}"

                    except KeyError as e:
                        logger.error(
                            f"variable id {variable} not found in GEFS lexicon"
                        )
                        raise e

                    # Get byte range from index
                    byte_offset = None
                    byte_length = None
                    for key, value in index_files[product].items():
                        if gefs_key in key:
                            byte_offset = value[0]
                            byte_length = (
                                None if value[1] < 0 else value[1]
                            )  # Negative means to eof
                            break

                    if byte_offset is None:
                        logger.warning(
                            f"Variable {v} not found in index file for time {t} at {lt}, values will be unset"
                        )
                        continue

                    tasks.append(
                        GEFSAsyncTask(
                            data_array_indices=(i, j, k),
                            gefs_file_uri=self._grib_uri(t, lt, product),
                            gefs_byte_offset=byte_offset,
                            gefs_byte_length=byte_length,
                            gefs_modifier=modifier,
                        )
                    )
        return tasks

    async def fetch_wrapper(
        self,
        task: GEFSAsyncTask,
        xr_array: xr.DataArray,
    ) -> xr.DataArray:
        """Small wrapper to pack arrays into the DataArray"""
        out = await self.fetch_array(
            task.gefs_file_uri,
            task.gefs_byte_offset,
            task.gefs_byte_length,
            task.gefs_modifier,
        )
        i, j, k = task.data_array_indices
        xr_array[i, j, k] = out

    async def fetch_array(
        self,
        grib_uri: str,
        byte_offset: int,
        byte_length: int | None,
        modifier: Callable,
    ) -> np.ndarray:
        """Fetch GEFS data array. This will first fetch the index file to get byte range
        of the needed data, fetch the respective grib files and lastly combining grib
        files into single data array.

        Parameters
        ----------
        time : datetime
            Date time to fetch
        lead_time : timedelta
            Forecast lead time to fetch
        variables : list[str]
            Variables to fetch

        Returns
        -------
        xr.DataArray
            FS data array for given time and lead time
        """
        logger.debug(f"Fetching GEFS grib file: {grib_uri} {byte_offset}-{byte_length}")
        # Download the grib file to cache
        grib_file = await self._fetch_remote_file(
            grib_uri,
            byte_offset=byte_offset,
            byte_length=byte_length,
        )
        # Open into xarray data-array
        da = xr.open_dataarray(
            grib_file, engine="cfgrib", backend_kwargs={"indexpath": ""}
        )
        return modifier(da.values)

    def _validate_time(self, times: list[datetime]) -> None:
        """Verify if date time is valid for GEFS based on offline knowledge

        Parameters
        ----------
        times : list[datetime]
            list of date times to fetch data
        """
        for time in times:
            if not (time - datetime(1900, 1, 1)).total_seconds() % 21600 == 0:
                raise ValueError(
                    f"Requested date time {time} needs to be 6 hour interval for GEFS"
                )
            # Brute forces checked this, older dates missing data
            # Open an issue if incorrect, may of gotten updated
            if time < datetime(year=2020, month=9, day=23):
                raise ValueError(
                    f"Requested date time {time} needs to be after Sept 23rd, 2020 for GEFS"
                )

    @classmethod
    def _validate_leadtime(cls, lead_times: list[timedelta]) -> None:
        """Verify if lead time is valid for GEFS based on offline knowledge"""
        for delta in lead_times:
            # To update search "gefs." at https://noaa-gefs-pds.s3.amazonaws.com/index.html
            hours = int(delta.total_seconds() // 3600)
            if hours > 384 or hours < 0:
                raise ValueError(
                    f"Requested lead time {delta} can only be a max of 384 hours for GEFS"
                )

            # 3-hours supported for first 10 days
            if delta.total_seconds() // 3600 <= 240:
                if not delta.total_seconds() % 10800 == 0:
                    raise ValueError(
                        f"Requested lead time {delta} needs to be 3 hour interval for first 10 days in GEFS"
                    )
            # 6 hours for rest
            else:
                if not delta.total_seconds() % 21600 == 0:
                    raise ValueError(
                        f"Requested lead time {delta} needs to be 6 hour interval for last 6 days in GEFS"
                    )

    async def _fetch_index(self, index_uri: str) -> dict[str, tuple[int, int]]:
        """Fetch GEFS atmospheric index file

        Parameters
        ----------
        index_uri : str
            URI to grib index file to download

        Returns
        -------
        dict[str, tuple[int, int]]
            Dictionary of GEFS vairables (byte offset, byte length)
        """
        # Grab index file
        index_file = await self._fetch_remote_file(index_uri)
        with open(index_file) as file:
            index_lines = [line.rstrip() for line in file]
        # Add dummy variable at end of file with max offset so algo below works
        index_lines.append("xx:-1:d=xx:NULL:NULL:NULL:NULL")
        index_table: dict[str, tuple[int, int]] = {}

        index_table = {}
        for i, line in enumerate(index_lines[:-1]):
            lsplit = line.split(":")
            if len(lsplit) < 7:
                continue

            nlsplit = index_lines[i + 1].split(":")
            byte_length = int(nlsplit[1]) - int(lsplit[1])
            byte_offset = int(lsplit[1])
            key = f"{lsplit[0]}::{lsplit[3]}::{lsplit[4]}::{lsplit[5]}"
            if byte_length > self.MAX_BYTE_SIZE:
                raise ValueError(
                    f"Byte length, {byte_length}, of variable {key} larger than safe threshold of {self.MAX_BYTE_SIZE}"
                )

            index_table[key] = (byte_offset, byte_length)

        return index_table

    async def _fetch_remote_file(
        self, path: str, byte_offset: int = 0, byte_length: int | None = None
    ) -> str:
        """Fetches remote file into cache"""
        sha = hashlib.sha256((path + str(byte_offset)).encode())
        filename = sha.hexdigest()
        cache_path = os.path.join(self.cache, filename)

        if not pathlib.Path(cache_path).is_file():
            if self.fs.async_impl:
                if byte_length:
                    byte_length = int(byte_offset + byte_length)
                data = await self.fs._cat_file(path, start=byte_offset, end=byte_length)
            else:
                data = await asyncio.to_thread(
                    self.fs.read_block, path, offset=byte_offset, length=byte_length
                )
            with open(cache_path, "wb") as file:
                await asyncio.to_thread(file.write, data)

        return cache_path

    def _grib_uri(
        self, time: datetime, lead_time: timedelta, product: str = "pgrb2a"
    ) -> str:
        """Generates the URI for GEFS grib files"""
        lead_hour = int(lead_time.total_seconds() // 3600)
        file_name = f"gefs.{time.year}{time.month:0>2}{time.day:0>2}/{time.hour:0>2}"
        p_res = self._product_resolution.replace("0", "")
        file_name = os.path.join(file_name, f"{self._product_class}/{product}{p_res}")
        file_name = os.path.join(
            file_name,
            f"{self._member}.t{time.hour:0>2}z.{product}.{self._product_resolution}.f{lead_hour:03d}",
        )
        return os.path.join(self.GEFS_BUCKET_NAME, file_name)

    def _grib_index_uri(
        self, time: datetime, lead_time: timedelta, product: str = "pgrb2a"
    ) -> str:
        """Generates the URI for GEFS index grib files"""
        # https://noaa-gefs-pds.s3.amazonaws.com/index.html#gefs.20250513/00/atmos/pgrb2ap5/
        lead_hour = int(lead_time.total_seconds() // 3600)
        file_name = f"gefs.{time.year}{time.month:0>2}{time.day:0>2}/{time.hour:0>2}"
        p_res = self._product_resolution.replace("0", "")
        file_name = os.path.join(file_name, f"{self._product_class}/{product}{p_res}")
        file_name = os.path.join(
            file_name,
            f"{self._member}.t{time.hour:0>2}z.{product}.{self._product_resolution}.f{lead_hour:03d}.idx",
        )
        return os.path.join(self.GEFS_BUCKET_NAME, file_name)

    @property
    def cache(self) -> str:
        """Return appropriate cache location."""
        cache_location = os.path.join(datasource_cache_root(), "gefs")
        if not self._cache:
            cache_location = os.path.join(cache_location, "tmp_gefs")
        return cache_location

    @classmethod
    def available(
        cls,
        time: datetime | np.datetime64,
    ) -> bool:
        """Checks if given date time is avaliable in the GEFS object store. Uses S3
        store

        Parameters
        ----------
        time : datetime | np.datetime64
            Date time to access

        Returns
        -------
        bool
            If date time is avaiable
        """
        if isinstance(time, np.datetime64):  # np.datetime64 -> datetime
            _unix = np.datetime64(0, "s")
            _ds = np.timedelta64(1, "s")
            time = datetime.fromtimestamp((time - _unix) / _ds, timezone.utc)

        fs = s3fs.S3FileSystem(anon=True)
        # Object store directory for given time
        # Just picking the first variable to look for
        file_name = f"gefs.{time.year}{time.month:0>2}{time.day:0>2}/{time.hour:0>2}/atmos/{cls.GEFS_CHECK_PRODUCT}/"
        s3_uri = f"s3://{cls.GEFS_BUCKET_NAME}/{file_name}"
        exists = fs.exists(s3_uri)

        return exists


class GEFS_FX_721x1440(GEFS_FX):
    """The Global Ensemble Forecast System (GEFS) forecast source is a 30 member
    ensemble forecast provided on an 0.25 degree equirectangular grid. GEFS is a
    weather forecast model developed by  National Centers for Environmental Prediction
    (NCEP). This data source provides the select variables of GEFS served on a higher
    resolution grid t 6-hour intervals spanning from Sept 23rd 2020 to present date.
    Each forecast provides 3-hourly predictions up to 10 days (240 hours) and 6 hourly
    predictions for another 6 days (384 hours).

    Parameters
    ----------
    member : str, optional
        GEFS member. Options are: control gec00 (control), gepNN (forecast member NN,
        e.g. gep01, gep02,...), by default "gec00"
    max_workers : int, optional
        Max works in async io thread pool. Only applied when using sync call function
        and will modify the default async loop if one exists, by default 24
    cache : bool, optional
        Cache data source on local memory, by default True
    verbose : bool, optional
        Print download progress, by default True
    async_timeout : int, optional
        Time in sec after which download will be cancelled if not finished successfully,
        by default 600

    Warning
    -------
    This is a remote data source and can potentially download a large amount of data
    to your local machine for large requests.

    Warning
    -------
    Some variables in the GEFS lexicon may not be available at lead time 0. Consult GEFS
    documentation for additional information.

    Note
    ----
    NCEP only provides a small subset of variables on the higher resoluton 0.25 degree
    grid. For a larger selection, use the standard GEFS data source.

    Note
    ----
    Additional information on the data repository can be referenced here:

    - https://registry.opendata.aws/noaa-gefs/
    - https://www.ncei.noaa.gov/products/weather-climate-models/global-ensemble-forecast
    - https://www.nco.ncep.noaa.gov/pmb/products/gens/
    """

    GEFS_LAT = np.linspace(90, -90, 721)
    GEFS_LON = np.linspace(0, 359.75, 1440)

    GEFS_CHECK_PRODUCT = "pgrb2sp25"

    def __init__(
        self,
        member: str = "gec00",
        max_workers: int = 24,
        cache: bool = True,
        verbose: bool = True,
        async_timeout: int = 600,
    ):
        super().__init__(
            member=member,
            max_workers=max_workers,
            cache=cache,
            verbose=verbose,
            async_timeout=async_timeout,
        )
        self._product_resolution = "0p25"
        self.lexicon = GEFSLexiconSel  # type: ignore

    def __call__(
        self,
        time: datetime | list[datetime] | TimeArray,
        lead_time: timedelta | list[timedelta] | LeadTimeArray,
        variable: str | list[str] | VariableArray,
    ) -> xr.DataArray:
        """Retrieve GEFS forecast data

        Parameters
        ----------
        time : datetime | list[datetime] | TimeArray
            Timestamps to return data for (UTC).
        lead_time: timedelta | list[timedelta] | LeadTimeArray
            Forecast lead times to fetch.
        variable : str | list[str] | VariableArray
            String, list of strings or array of strings that refer to variables to
            return. Must be in the GEFS FX select lexicon.

        Returns
        -------
        xr.DataArray
            GEFS select forecast data array
        """
        return super().__call__(time, lead_time, variable)

    @classmethod
    def _validate_leadtime(cls, lead_times: list[timedelta]) -> None:
        """Verify if lead time is valid for GEFS based on offline knowledge"""
        for delta in lead_times:
            # To update search "gefs." at https://noaa-gefs-pds.s3.amazonaws.com/index.html
            hours = int(delta.total_seconds() // 3600)
            if hours > 240 or hours < 0:
                raise ValueError(
                    f"Requested lead time {delta} can only be a max of 240 hours for GEFS 0.25 degree data"
                )

            # 3-hours supported for first 10 days
            if delta.total_seconds() // 3600 <= 240:
                if not delta.total_seconds() % 10800 == 0:
                    raise ValueError(
                        f"Requested lead time {delta} needs to be 3 hour interval for first 10 days in GEFS 0.25 degree data"
                    )



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import zipfile
from collections import OrderedDict
from collections.abc import Generator, Iterator
from datetime import timedelta
from pathlib import Path

import numpy as np
import torch
import xarray

try:
    import physicsnemo
    from physicsnemo.utils.zenith_angle import cos_zenith_angle
except ImportError:
    physicsnemo = None
    cos_zenith_angle = None

from earth2studio.models.auto import AutoModelMixin, Package
from earth2studio.models.batch import batch_coords, batch_func
from earth2studio.models.px.base import PrognosticModel
from earth2studio.models.px.utils import PrognosticMixin
from earth2studio.utils import check_extra_imports, handshake_coords, handshake_dim
from earth2studio.utils.time import timearray_to_datetime
from earth2studio.utils.type import CoordSystem

VARIABLES = ["t850", "z1000", "z700", "z500", "z300", "tcwv", "t2m"]


@check_extra_imports("dlwp", [physicsnemo, cos_zenith_angle])
class DLWP(torch.nn.Module, AutoModelMixin, PrognosticMixin):
    """Deep learning weather prediction (DLWP)  prognostic model. This is a parsimonious
    global forecast model with a time-step size of 6 hours. The core model is a
    convolutional encoder-decoder trained on [64,64] cubed sphere data that has an input
    of 18 fields (2x7 atmos variables + 4 prescriptive) and outputs 14 fields (2x7 atmos
    variables). This implementation provides a wrapper that accepts [721,1440] lat-lon
    equirectangular grid of just the atmospheric varaibles as an input for better
    compatability with common data sources. Prescriptive fields are added inside the
    model wrapper.

    Note
    ----
    For more information about this model see:

    - https://agupubs.onlinelibrary.wiley.com/doi/epdf/10.1029/2021MS002502
    - https://catalog.ngc.nvidia.com/orgs/nvidia/teams/modulus/models/modulus_dlwp_cubesphere

    Parameters
    ----------
    core_model : torch.nn.Module
        Core cubed-sphere DLWP model.
    landsea_mask : torch.Tensor
        Land sea mask in cubed sphere form [6,64,64]
    orography : torch.Tensor
        Surface geopotential (orography) in cubed sphere form [6,64,64]
    latgrid : torch.Tensor
        Cubed sphere latitude coordinates [6,64,64]
    longrid : torch.Tensor
        Cubed sphere longitude coordinates [6,64,64]
    cubed_sphere_transform : torch.Tensor
        Sparse pytorch tensor to transform equirectangular fields to cubed sphere of
        size [24576, 1038240]
    cubed_sphere_inverse : torch.Tensor
        Sparse pytorch tensor to transform cubed sphere fields to equirectangular of
        size [1038240, 24576]
    center : torch.Tensor
        Model atmospheric variable center normalization tensor of size [1,7,1,1]
    scale : torch.Tensor
        Model atmospheric variable scale normalization tensor of size [1,7,1,1]
    """

    def __init__(
        self,
        core_model: torch.nn.Module,
        landsea_mask: torch.Tensor,
        orography: torch.Tensor,
        latgrid: torch.Tensor,
        longrid: torch.Tensor,
        cubed_sphere_transform: torch.Tensor,
        cubed_sphere_inverse: torch.Tensor,
        center: torch.Tensor,
        scale: torch.Tensor,
    ):

        super().__init__()
        self.model = core_model
        self.register_buffer("latgrid", latgrid)
        self.register_buffer("longrid", longrid)
        self.register_buffer("center", center)
        self.register_buffer("scale", scale)
        self.register_buffer("landsea_mask", landsea_mask.unsqueeze(0))
        self.register_buffer(
            "topographic_height",
            (orography.unsqueeze(0).unsqueeze(0) - 3.724e03) / 8.349e03,
        )
        self.register_buffer("M", cubed_sphere_transform.T)
        self.register_buffer("N", cubed_sphere_inverse)

    def input_coords(self) -> CoordSystem:
        """Input coordinate system of the prognostic model

        Returns
        -------
        CoordSystem
            Coordinate system dictionary
        """
        return OrderedDict(
            {
                "batch": np.empty(0),
                "time": np.empty(0),
                "lead_time": np.array(
                    [np.timedelta64(-6, "h"), np.timedelta64(0, "h")]
                ),
                "variable": np.array(VARIABLES),
                "lat": np.linspace(90, -90, 721),
                "lon": np.linspace(0, 360, 1440, endpoint=False),
            }
        )

    @batch_coords()
    def output_coords(self, input_coords: CoordSystem) -> CoordSystem:
        """Output coordinate system of the prognostic model

        Parameters
        ----------
        input_coords : CoordSystem
            Input coordinate system to transform into output_coords

        Returns
        -------
        CoordSystem
            Coordinate system dictionary
        """

        output_coords = OrderedDict(
            {
                "batch": np.empty(0),
                "time": np.empty(0),
                "lead_time": np.array([np.timedelta64(6, "h")]),
                "variable": np.array(VARIABLES),
                "lat": np.linspace(90, -90, 721),
                "lon": np.linspace(0, 360, 1440, endpoint=False),
            }
        )

        test_coords = input_coords.copy()
        test_coords["lead_time"] = (
            test_coords["lead_time"] - input_coords["lead_time"][-1]
        )
        target_input_coords = self.input_coords()
        for i, key in enumerate(target_input_coords):
            handshake_dim(test_coords, key, i)
            if key not in ["batch", "time"]:
                handshake_coords(test_coords, target_input_coords, key)

        # Normal forward pass of DLWP, this method returns two time-steps
        output_coords["batch"] = input_coords["batch"]
        output_coords["time"] = input_coords["time"]

        output_coords["lead_time"] = (
            input_coords["lead_time"][-1] + output_coords["lead_time"]
        )

        return output_coords

    @classmethod
    def load_default_package(cls) -> Package:
        """Default DLWP model package on NGC"""
        return Package(
            "ngc://models/nvidia/modulus/modulus_dlwp_cubesphere@v0.2",
            cache_options={
                "cache_storage": Package.default_cache("dlwp"),
                "same_names": True,
            },
        )

    @classmethod
    @check_extra_imports("dlwp", [physicsnemo, cos_zenith_angle])
    def load_model(
        cls,
        package: Package,
    ) -> PrognosticModel:
        """Load prognostic from package"""
        # Ghetto at the moment because NGC files are zipped. This will download zip and
        # unpack them then give the cached folder location from which we can then
        # access the needed files.
        dlwp_zip = Path(package.resolve("dlwp_cubesphere.zip"))
        # Have to manually unzip here. Should not zip checkpoints in the future
        with zipfile.ZipFile(dlwp_zip, "r") as zip_ref:
            zip_ref.extractall(dlwp_zip.parent)

        lsm = torch.Tensor(
            xarray.open_dataset(
                str(dlwp_zip.parent / Path("dlwp/land_sea_mask_rs_cs.nc"))
            )["lsm"].values
        )
        topographic_height = torch.Tensor(
            xarray.open_dataset(
                str(dlwp_zip.parent / Path("dlwp/geopotential_rs_cs.nc"))
            )["z"].values
        )
        latlon_grids = xarray.open_dataset(
            str(dlwp_zip.parent / Path("dlwp/latlon_grid_field_rs_cs.nc"))
        )
        latgrid = torch.Tensor(latlon_grids["latgrid"].values)
        longrid = torch.Tensor(latlon_grids["longrid"].values)
        # load maps
        input_map_wts = xarray.open_dataset(
            str(dlwp_zip.parent / Path("dlwp/map_LL721x1440_CS64.nc"))
        )
        output_map_wts = xarray.open_dataset(
            str(dlwp_zip.parent / Path("dlwp/map_CS64_LL721x1440.nc"))
        )

        i = input_map_wts.row.values - 1
        j = input_map_wts.col.values - 1
        data = input_map_wts.S.values
        cubed_sphere_transform = torch.sparse_coo_tensor(
            np.array((i, j)), data, dtype=torch.float
        )

        i = output_map_wts.row.values - 1
        j = output_map_wts.col.values - 1
        data = output_map_wts.S.values
        cubed_sphere_inverse = torch.sparse_coo_tensor(
            np.array((i, j)), data, dtype=torch.float
        )

        core_model = physicsnemo.Module.from_checkpoint(
            str(dlwp_zip.parent / Path("dlwp/dlwp.mdlus"))
        )

        center = torch.Tensor(
            np.load(str(dlwp_zip.parent / Path("dlwp/global_means.npy")))
        )
        scale = torch.Tensor(
            np.load(str(dlwp_zip.parent / Path("dlwp/global_stds.npy")))
        )

        return cls(
            core_model,
            landsea_mask=lsm,
            orography=topographic_height,
            latgrid=latgrid,
            longrid=longrid,
            cubed_sphere_transform=cubed_sphere_transform,
            cubed_sphere_inverse=cubed_sphere_inverse,
            center=center,
            scale=scale,
        )

    def to_cubedsphere(self, x: torch.Tensor) -> torch.Tensor:
        """[721,1440] eqr to [6,64,64] cs"""
        x = x.reshape(*x.shape[:-2], -1) @ self.M
        x = x.reshape(*x.shape[:-1], 6, 64, 64)
        return x

    def to_equirectangular(self, x: torch.Tensor) -> torch.Tensor:
        """[6,64,64] cs to [721,1440] eqr"""
        input_shape = x.shape[:-3]
        x = (self.N @ x.reshape(-1, 6 * 64 * 64).T).T
        x = x.reshape(*input_shape, 721, 1440)
        return x

    def get_cosine_zenith_fields(
        self, times: np.array, lead_time: timedelta, device: torch.device | str = "cuda"
    ) -> torch.Tensor:
        """Creates cosine zenith fields for input time array"""
        output = []
        for time in timearray_to_datetime(times):
            uvcossza = cos_zenith_angle(
                time + lead_time,
                self.longrid.cpu(),
                self.latgrid.cpu(),
            )
            # Normalize
            uvcossza = torch.Tensor(uvcossza).to(device)
            uvcossza = torch.clamp(uvcossza, min=0) - 1.0 / np.pi
            output.append(uvcossza)
        return torch.stack(output, axis=0)

    def _prepare_input(self, input: torch.Tensor, coords: CoordSystem) -> torch.Tensor:
        """Prepares input cubed sphere tensor by adding land sea mask, uvcossza and
        orography fields to input atmospheric ([14,6,64,64] -> [18,6,64,64])
        """
        # Compress batch dim into time
        time_array = np.tile(coords["time"], input.shape[0])
        input = input.view(-1, *input.shape[2:])

        uvcossza_6 = self.get_cosine_zenith_fields(
            time_array, timedelta(hours=-6), input.device
        ).unsqueeze(1)
        uvcossza_0 = self.get_cosine_zenith_fields(
            time_array, timedelta(hours=0), input.device
        ).unsqueeze(1)
        x = torch.cat([input[:, 0], uvcossza_6, input[:, 1], uvcossza_0], dim=1)

        input = torch.cat(
            (
                x,
                self.landsea_mask.repeat(x.shape[0], 1, 1, 1, 1),
                self.topographic_height.repeat(x.shape[0], 1, 1, 1, 1),
            ),
            dim=1,
        )
        return input

    def _prepare_output(
        self, output: torch.Tensor, coords: CoordSystem
    ) -> torch.Tensor:
        output = torch.split(output, output.shape[1] // 2, dim=1)
        # Add lead time dimension back in
        output = torch.stack(output, dim=1)
        # Add batch dimension back in
        output = output.view(-1, coords["time"].shape[0], *output.shape[1:])
        return output

    @torch.inference_mode()
    def _forward(
        self,
        x: torch.Tensor,
        coords: CoordSystem,
    ) -> torch.Tensor:

        center = self.center.unsqueeze(-1)
        scale = self.scale.unsqueeze(-1)

        x = (x - center) / scale
        x = self._prepare_input(x, coords)
        x = self.model(x)
        x = self._prepare_output(x, coords)
        x = scale * x + center
        return x

    @batch_func()
    def __call__(
        self,
        x: torch.Tensor,
        coords: CoordSystem,
    ) -> tuple[torch.Tensor, CoordSystem]:
        """Runs prognostic model 1 step.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor
        coords : CoordSystem
            Input coordinate system

        Returns
        -------
        tuple[torch.Tensor, CoordSystem]
            Output tensor and coordinate system 6 hours in the future
        """

        output_coords = self.output_coords(coords)

        x = self.to_cubedsphere(x)
        x = self._forward(x, coords)
        x = self.to_equirectangular(x)

        return x[:, :, :1], output_coords

    @batch_func()
    def _default_generator(
        self, x: torch.Tensor, coords: CoordSystem
    ) -> Generator[tuple[torch.Tensor, CoordSystem], None, None]:

        coords = coords.copy()
        self.output_coords(coords)

        coords_out = coords.copy()
        coords_out["lead_time"] = coords["lead_time"][1:]
        yield x[:, :, 1:], coords_out

        x = self.to_cubedsphere(x)
        while True:
            # Front hook
            x, coords = self.front_hook(x, coords)

            # Forward pass
            x = self._forward(x, coords)
            coords["lead_time"] = (
                coords["lead_time"]
                + 2 * self.output_coords(self.input_coords())["lead_time"]
            )
            x = x.clone()

            # Rear hook for first predicted step
            coords_out = coords.copy()
            coords_out["lead_time"] = coords["lead_time"][0]
            x[:, :, :1], coords_out = self.rear_hook(x[:, :, :1], coords_out)

            # Output first predicted step
            out = self.to_equirectangular(x[:, :, :1])
            yield out, coords_out

            # Rear hook for second predicted step
            coords_out["lead_time"] = coords["lead_time"][-1]
            x[:, :, 1:], coords_out = self.rear_hook(x[:, :, 1:], coords_out)
            out = self.to_equirectangular(x[:, :, 1:])
            yield out, coords_out

    def create_iterator(
        self, x: torch.Tensor, coords: CoordSystem
    ) -> Iterator[tuple[torch.Tensor, CoordSystem]]:
        """Creates a iterator which can be used to perform time-integration of the
        prognostic model. Will return the initial condition first (0th step).

        Parameters
        ----------
        x : torch.Tensor
            Input tensor
        coords : CoordSystem
            Input coordinate system


        Yields
        ------
        Iterator[tuple[torch.Tensor, CoordSystem]]
            Iterator that generates time-steps of the prognostic model container the
            output data tensor and coordinate system dictionary.
        """
        yield from self._default_generator(x, coords)



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
from collections.abc import Generator, Iterator
from pathlib import Path

import numpy as np
import torch
import xarray as xr

try:
    import earth2grid
    from omegaconf import OmegaConf
    from physicsnemo.models import Module
    from physicsnemo.utils.insolation import insolation
except ImportError:
    Module = None
    insolation = None
    OmegaConf = None
    earth2grid = None
from earth2studio.models.auto import AutoModelMixin, Package
from earth2studio.models.batch import batch_coords, batch_func
from earth2studio.models.px.base import PrognosticModel
from earth2studio.models.px.utils import PrognosticMixin
from earth2studio.utils import check_extra_imports, handshake_coords, handshake_dim
from earth2studio.utils.type import CoordSystem

_ATMOS_VARIABLES = [
    "z500",
    "tau300-700",
    "z1000",
    "t2m",
    "tcwv",
    "t850",
    "z250",
    "ws10m",
]
_OCEAN_VARIABLES = ["sst"]
_ATMOS_COUPLING_VARIABLES = ["sst"]
_OCEAN_COUPLING_VARIABLES = ["z1000", "ws10m"]

_ATMOS_INPUT_TIMES = np.array([-18, -12, -6, 0], dtype="timedelta64[h]")
_OCEAN_INPUT_TIMES = np.array([-48, 0], dtype="timedelta64[h]")

_ATMOS_OUTPUT_TIMES = np.array(
    [6, 12, 18, 24, 30, 36, 42, 48, 54, 60, 66, 72, 78, 84, 90, 96],
    dtype="timedelta64[h]",
)
_OCEAN_OUTPUT_TIMES = np.array([48, 96], dtype="timedelta64[h]")


@check_extra_imports("dlesym", [OmegaConf, Module, insolation])
class DLESyM(torch.nn.Module, AutoModelMixin, PrognosticMixin):
    """DLESyM-V1-ERA5 prognostic model. This is an ensemble forecast model for
    global earth system modeling. This model includes an atmosphere and ocean
    component, using atmospheric variables as well as the sea-surface temperature
    on a HEALPix nside=64 (approximately 1 degree) resolution grid. The model
    architecture is a U-Net with padding operations modified to support using
    the HEALPix grid. Because the atmosphere and ocean models are predicted at
    different times, not all entries in the output tensor are valid. As a result,
    we provide convenience methods for retrieving the valid atmospheric and
    oceanic outputs.

    Example
    -------
    .. highlight:: python
    .. code-block:: python

        pkg = DLESyM.load_default_package()
        model = DLESyM.load_model(pkg)

        # Create iterator
        iterator = model.create_iterator(x, coords)

        for step, (x, coords) in enumerate(iterator):
            if step > 0:
                # Valid atmos and ocean predictions with their respective coordinates extracted below
                atmos_outputs, atmos_coords = model.retrieve_valid_atmos_outputs(x, coords)
                ocean_outputs, ocean_coords = model.retrieve_valid_ocean_outputs(x, coords)
                ...

    Note
    ----
    For more information about this model see:

    - https://arxiv.org/abs/2409.16247
    - https://arxiv.org/abs/2311.06253v2

    For more information about the HEALPix grid see:

    - https://github.com/NVlabs/earth2grid

    Parameters
    ----------
    atmos_model : torch.nn.Module
        Atmosphere model
    ocean_model : torch.nn.Module
        Ocean model
    hpx_lat : np.ndarray
        HEALPix latitude coordinates, shape (12, nside, nside)
    hpx_lon : np.ndarray
        HEALPix longitude coordinates, shape (12, nside, nside)
    nside : int
        HEALPix nside
    center : np.ndarray
        Means of the input data, shape (1, 1, 1, num_variables, 1, 1, 1)
    scale : np.ndarray
        Standard deviations of the input data, shape (1, 1, 1, num_variables, 1, 1, 1)
    atmos_constants : np.ndarray
        Constants for the atmosphere model, shape (12, num_atmos_constants, nside, nside)
    ocean_constants : np.ndarray
        Constants for the ocean model, shape (12, num_ocean_constants, nside, nside)
    atmos_input_times : np.ndarray
        Atmospheric input times, shape (num_atmos_input_times,)
    ocean_input_times : np.ndarray
        Ocean input times, shape (num_ocean_input_times,)
    atmos_output_times : np.ndarray
        Atmospheric output times, shape (num_atmos_output_times,)
    atmos_variables : list[str]
        Atmospheric variables
    ocean_variables : list[str]
        Ocean variables
    atmos_coupling_variables : list[str]
        Atmospheric coupling variables
    ocean_coupling_variables : list[str]
        Ocean coupling variables
    """

    def __init__(
        self,
        atmos_model: torch.nn.Module,
        ocean_model: torch.nn.Module,
        hpx_lat: np.ndarray,
        hpx_lon: np.ndarray,
        nside: int,
        center: np.ndarray,
        scale: np.ndarray,
        atmos_constants: np.ndarray,
        ocean_constants: np.ndarray,
        atmos_input_times: np.ndarray,
        ocean_input_times: np.ndarray,
        atmos_output_times: np.ndarray,
        ocean_output_times: np.ndarray,
        atmos_variables: list[str],
        ocean_variables: list[str],
        atmos_coupling_variables: list[str],
        ocean_coupling_variables: list[str],
    ):

        super().__init__()
        self.atmos_model = atmos_model.eval()
        self.ocean_model = ocean_model.eval()

        self.register_buffer("center", torch.from_numpy(center).to(dtype=torch.float32))
        self.register_buffer("scale", torch.from_numpy(scale).to(dtype=torch.float32))
        self.register_buffer(
            "atmos_constants", torch.from_numpy(atmos_constants).to(dtype=torch.float32)
        )
        self.register_buffer(
            "ocean_constants", torch.from_numpy(ocean_constants).to(dtype=torch.float32)
        )

        self.hpx_lat = hpx_lat
        self.hpx_lon = hpx_lon
        self.nside = nside
        self.atmos_variables = atmos_variables
        self.ocean_variables = ocean_variables
        self.atmos_coupling_variables = atmos_coupling_variables
        self.ocean_coupling_variables = ocean_coupling_variables

        # Validate the input and output times
        for name, times in zip(
            [
                "atmos_input_times",
                "ocean_input_times",
                "atmos_output_times",
                "ocean_output_times",
            ],
            [
                atmos_input_times,
                ocean_input_times,
                atmos_output_times,
                ocean_output_times,
            ],
        ):
            if not np.issubdtype(times.dtype, np.timedelta64):
                raise ValueError(
                    f"Input and output times must be of type 'timedelta64', got {times.dtype} for {name}"
                )
        if ocean_input_times[0] > atmos_input_times[0]:
            raise ValueError(
                f"Ocean input time must be equal to or before atmos input time, got {ocean_input_times[0]} and {atmos_input_times[0]}"
            )
        if len(atmos_input_times) < len(ocean_input_times):
            raise ValueError(
                f"Atmos input times must be equal to or longer than ocean input times, got {len(atmos_input_times)} and {len(ocean_input_times)}"
            )
        if len(atmos_output_times) < len(ocean_output_times):
            raise ValueError(
                f"Atmos output times must be equal to or longer than ocean output times, got {len(atmos_output_times)} and {len(ocean_output_times)}"
            )

        self.atmos_input_times = atmos_input_times
        self.ocean_input_times = ocean_input_times
        # Full input times are a merge of the atmos and ocean input times:
        #  - Need to go as far back as -48h for the first ocean input time
        #  - Need to use the same temporal resolution (6h) as the atmos input times
        self.full_input_times = np.arange(
            self.ocean_input_times[0],
            self.atmos_input_times[-1] + 1,
            atmos_input_times[1] - atmos_input_times[0],
        )
        self.atmos_output_times = atmos_output_times
        self.ocean_output_times = ocean_output_times

        # Setup the insolation and coupling times from the input and output times
        self.atmos_sol_times = np.concatenate(
            [self.atmos_input_times, self.atmos_output_times]
        )
        self.ocean_sol_times = np.concatenate(
            [self.ocean_input_times, self.ocean_output_times]
        )
        n_atmos_steps = 1 + max(
            self.atmos_model.output_time_dim // self.atmos_model.input_time_dim, 1
        )
        self.atmos_coupling_times = np.array(
            [self.atmos_input_times[-1]] * n_atmos_steps
        )
        self.ocean_coupling_times = self.atmos_output_times

        # Setup the lead time indices for [atmos, ocean] [input, coupled input, output]
        in_coords = self.input_coords()
        out_coords = self.output_coords(in_coords)
        self.atmos_input_lt_idx = [
            list(in_coords["lead_time"]).index(t) for t in self.atmos_input_times
        ]
        self.ocean_input_lt_idx = [
            list(in_coords["lead_time"]).index(t) for t in self.ocean_input_times
        ]
        self.atmos_coupled_input_lt_idx = [
            list(in_coords["lead_time"]).index(t) for t in self.atmos_coupling_times
        ]
        self.ocean_coupled_input_lt_idx = [
            list(out_coords["lead_time"]).index(t) for t in self.ocean_coupling_times
        ]
        self.atmos_output_lt_idx = [
            list(out_coords["lead_time"]).index(t) for t in self.atmos_output_times
        ]
        self.ocean_output_lt_idx = [
            list(out_coords["lead_time"]).index(t) for t in self.ocean_output_times
        ]

        # Setup the variable indices for [atmos, ocean]
        self.atmos_var_idx = [
            list(out_coords["variable"]).index(var) for var in self.atmos_variables
        ]
        self.ocean_var_idx = [
            list(out_coords["variable"]).index(var) for var in self.ocean_variables
        ]
        self.atmos_coupling_var_idx = [
            list(out_coords["variable"]).index(var)
            for var in self.atmos_coupling_variables
        ]
        self.ocean_coupling_var_idx = [
            list(out_coords["variable"]).index(var)
            for var in self.ocean_coupling_variables
        ]

    def input_coords(self) -> CoordSystem:
        """Input coordinate system of the prognostic model

        Returns
        -------
        CoordSystem
            Coordinate system dictionary
        """
        return OrderedDict(
            {
                "batch": np.empty(0),
                "time": np.empty(0),
                "lead_time": self.full_input_times,
                "variable": np.array(self.atmos_variables + self.ocean_variables),
                "face": np.arange(12),
                "height": np.arange(self.nside),
                "width": np.arange(self.nside),
            }
        )

    @batch_coords()
    def output_coords(self, input_coords: CoordSystem) -> CoordSystem:
        """Output coordinate system of the prognostic model

        Parameters
        ----------
        input_coords : CoordSystem
            Input coordinate system to transform into output_coords

        Returns
        -------
        CoordSystem
            Coordinate system dictionary
        """

        output_coords = OrderedDict(
            {
                "batch": np.empty(0),
                "time": np.empty(0),
                "lead_time": self.atmos_output_times,  # atmos model has the finer temporal resolution over output lead times
                "variable": np.array(self.atmos_variables + self.ocean_variables),
                "face": np.arange(12),
                "height": np.arange(self.nside),
                "width": np.arange(self.nside),
            }
        )

        test_coords = input_coords.copy()
        test_coords["lead_time"] = (
            test_coords["lead_time"] - input_coords["lead_time"][-1]
        )

        target_input_coords = self.input_coords()
        for i, key in enumerate(target_input_coords):
            if key not in ["batch", "time"]:
                handshake_dim(test_coords, key, i)
                handshake_coords(test_coords, target_input_coords, key)

        output_coords["batch"] = input_coords["batch"]
        output_coords["time"] = input_coords["time"]
        output_coords["lead_time"] = (
            input_coords["lead_time"][-1] + output_coords["lead_time"]
        )

        return output_coords

    @classmethod
    def load_default_package(cls) -> Package:
        """Default DLESyM model package on NGC"""
        package = Package(
            "ngc://models/nvidia/earth-2/dlesym-v1-era5@1.0.1",
            cache_options={
                "cache_storage": Package.default_cache("dlesym"),
                "same_names": True,
            },
        )
        return package

    @classmethod
    @check_extra_imports("dlesym", [Module, OmegaConf])
    def load_model(
        cls,
        package: Package,
        atmos_model_idx: int = 0,
        ocean_model_idx: int = 0,
    ) -> PrognosticModel:
        """Load prognostic from package

        Parameters
        ----------
        package : Package
            Package to load model from
        atmos_model_idx : int, optional
            Index of atmos model weights in package to load, by default 0
        ocean_model_idx : int, optional
            Index of ocean model weights in package to load, by default 0

        Returns
        -------
        PrognosticModel
            Prognostic model
        """

        cfg_file = Path(package.resolve("config.yaml"))
        cfg = OmegaConf.load(cfg_file)
        nside = cfg.data.nside

        atmos_model_ckpt = package.resolve(
            cfg.models.atmos_model_checkpoints[atmos_model_idx]
        )
        ocean_model_ckpt = package.resolve(
            cfg.models.ocean_model_checkpoints[ocean_model_idx]
        )

        atmos_model = Module.from_checkpoint(atmos_model_ckpt)
        ocean_model = Module.from_checkpoint(ocean_model_ckpt)
        atmos_model.output_time_dim = len(cfg.io.atmos_output_times)
        ocean_model.output_time_dim = len(cfg.io.ocean_output_times)

        # Normalization constants
        ctr = np.array(
            [cfg.data.scaling[var]["mean"] for var in cfg.io.atmos_variables]
            + [cfg.data.scaling[var]["mean"] for var in cfg.io.ocean_variables]
        )
        scl = np.array(
            [cfg.data.scaling[var]["std"] for var in cfg.io.atmos_variables]
            + [cfg.data.scaling[var]["std"] for var in cfg.io.ocean_variables]
        )
        center = ctr[None, None, None, :, None, None, None]
        scale = scl[None, None, None, :, None, None, None]

        # Constant fields
        hpx_lat = np.load(package.resolve("hpx_lat.npy"))
        hpx_lon = np.load(package.resolve("hpx_lon.npy"))
        atmos_constants = np.stack(
            [
                np.load(package.resolve(f"{const}.npy"))
                for const in cfg.io.atmos_constants
            ],
            axis=1,
        )
        ocean_constants = np.stack(
            [
                np.load(package.resolve(f"{const}.npy"))
                for const in cfg.io.ocean_constants
            ],
            axis=1,
        )

        return cls(
            atmos_model,
            ocean_model,
            center=center,
            scale=scale,
            atmos_constants=atmos_constants,
            ocean_constants=ocean_constants,
            hpx_lat=hpx_lat,
            hpx_lon=hpx_lon,
            nside=nside,
            atmos_input_times=np.array(
                cfg.io.atmos_input_times, dtype="timedelta64[h]"
            ),
            ocean_input_times=np.array(
                cfg.io.ocean_input_times, dtype="timedelta64[h]"
            ),
            atmos_output_times=np.array(
                cfg.io.atmos_output_times, dtype="timedelta64[h]"
            ),
            ocean_output_times=np.array(
                cfg.io.ocean_output_times, dtype="timedelta64[h]"
            ),
            atmos_variables=cfg.io.atmos_variables,
            ocean_variables=cfg.io.ocean_variables,
            atmos_coupling_variables=cfg.io.atmos_coupling_variables,
            ocean_coupling_variables=cfg.io.ocean_coupling_variables,
        )

    def prepare_input_data(
        self, x: torch.Tensor, coords: CoordSystem
    ) -> tuple[list[torch.Tensor], list[torch.Tensor]]:
        """Prepare input data for the atmos and ocean models.
        From the data in `x`, we will build a list of tensors for each model.
        Assumes `x` is a tensor of shape (batch, time, lead_time, variable, face, height, width).
        The list order is [state, insolation, constants, coupled_inputs].
        Models expect input data in the following shapes:
            state: (batch, face, lead_time, variable, height, width)
            insolation: (batch, face, lead_time, variable(=1), height, width)
            constants: (face, variable, height, width)
            coupled_inputs: (lead_time, batch, variable, face, height, width)

        Parameters
        ----------
        x : torch.Tensor
            Input data
        coords : CoordSystem
            Input coordinates

        Returns
        -------
        tuple[List[torch.Tensor], List[torch.Tensor]]
            Prepared input data for atmos and ocean models, respectively
        """

        if x.ndim != 7:
            raise ValueError(
                f"DLESyM input data must be of shape (batch, time, lead_time, variable, face, height, width), got {x.shape} for coords {coords.keys()}"
            )

        # Flatten the batch and time dimensions
        # Before flattening, check if we have multiple batch elements
        if x.shape[0] > 1:
            # Roll the times for each batch element into one array for computing insolation
            stacked_times = np.concatenate([coords["time"]] * x.shape[0], axis=0)
        else:
            stacked_times = coords["time"]

        x = x.reshape(-1, *x.shape[2:])

        # Atmos inputs: state, insolation, constants, coupled inputs
        atmos_state = x[:, self.atmos_input_lt_idx][
            ..., self.atmos_var_idx, :, :, :
        ].permute(0, 3, 1, 2, 4, 5)
        atmos_insolation = self._make_insolation_tensor(
            anchor_times=stacked_times,
            timedeltas=self.atmos_sol_times + coords["lead_time"][-1],
        )
        atmos_coupling = self._make_atmos_coupling(x, coords)
        atmos_inputs = list(
            map(
                lambda y: y.to(x.device, x.dtype),
                [atmos_state, atmos_insolation, self.atmos_constants, atmos_coupling],
            )
        )

        # Ocean inputs: state, insolation, constants, coupled inputs
        # Coupling is not set as the ocean coupling comes after the atmos model forward pass
        ocean_state = x[:, self.ocean_input_lt_idx][
            ..., self.ocean_var_idx, :, :, :
        ].permute(0, 3, 1, 2, 4, 5)
        ocean_insolation = self._make_insolation_tensor(
            anchor_times=stacked_times,
            timedeltas=self.ocean_sol_times + coords["lead_time"][-1],
        )
        ocean_inputs = list(
            map(
                lambda y: y.to(x.device, x.dtype),
                [ocean_state, ocean_insolation, self.ocean_constants],
            )
        )

        return atmos_inputs, ocean_inputs

    def prepare_output_data(
        self,
        atmos_outputs: torch.Tensor,
        ocean_outputs: torch.Tensor,
        coords: CoordSystem,
    ) -> torch.Tensor:
        """Prepare output data for the atmos and ocean models.
        From the data in `atmos_outputs` and `ocean_outputs`, we will build a tensor of shape (batch, time, lead_time, variable, face, height, width).
        Assumes the atmosphere has the finest temporal resolution over output lead times.
        Expects input data to be of shape (batch * time, face, lead_time, variable, height, width).

        Parameters
        ----------
        atmos_outputs : torch.Tensor
            Atmos outputs
        ocean_outputs : torch.Tensor
            Ocean outputs
        coords : CoordSystem
            Input coordinates

        Returns
        -------
        output_data : torch.Tensor
            Output data
        """

        output_data = torch.empty(
            (
                len(coords["batch"]),
                len(coords["time"]),
                len(self.atmos_output_times),
                len(coords["variable"]),
                len(coords["face"]),
                len(coords["height"]),
                len(coords["width"]),
            ),
            device=atmos_outputs.device,
        )

        # Reorder the face dim to after variable dim, restore the batch and time dimensions
        atmos_outputs = atmos_outputs.permute(0, 2, 3, 1, 4, 5)
        ocean_outputs = ocean_outputs.permute(0, 2, 3, 1, 4, 5)
        atmos_outputs = atmos_outputs.reshape(
            len(coords["batch"]), len(coords["time"]), *atmos_outputs.shape[1:]
        )
        ocean_outputs = ocean_outputs.reshape(
            len(coords["batch"]), len(coords["time"]), *ocean_outputs.shape[1:]
        )

        output_data[:, :, :, self.atmos_var_idx, :, :, :] = atmos_outputs
        for src_idx, dst_idx in enumerate(self.ocean_var_idx):
            output_data[:, :, self.ocean_output_lt_idx, dst_idx, :, :, :] = (
                ocean_outputs[:, :, :, src_idx, :, :, :]
            )

        return output_data

    @check_extra_imports("dlesym", [insolation])
    def _make_insolation_tensor(
        self, anchor_times: np.ndarray, timedeltas: np.ndarray
    ) -> torch.Tensor:
        """Make insolation tensor from anchor times and timedeltas

        Parameters
        ----------
        anchor_times : np.ndarray
            Anchor times
        timedeltas : np.ndarray
            Timedeltas

        Returns
        -------
        torch.Tensor
            Insolation tensor of shape (1, face, lead_time, variable, height, width)
        """

        # Create insolation tensor, shape (len(anchor_times) * len(timedeltas), face, height, width)
        times_flattened = np.array(
            [[a + t for t in timedeltas] for a in anchor_times]
        ).flatten()
        sol = insolation(
            times_flattened,
            self.hpx_lat,
            self.hpx_lon,
        )
        t, f, h, w = sol.shape
        sol = torch.from_numpy(sol).view(len(anchor_times), len(timedeltas), 1, f, h, w)
        sol = sol.permute(0, 3, 1, 2, 4, 5)
        return sol

    def _make_atmos_coupling(
        self, x: torch.Tensor, coords: CoordSystem
    ) -> torch.Tensor:
        """Make atmos coupling tensor from input data
        Assumes `x` is a tensor of shape (batch, lead_time, variable, face, height, width).
        Also assumes `x` has been properly normalized.

        Parameters
        ----------
        x : torch.Tensor
            Input data
        coords : CoordSystem
            Input coordinates

        Returns
        -------
        torch.Tensor
            Atmos coupling tensor: shape (lead_time, batch, variable, face, height, width)
        """

        atmos_coupling = x[:, self.atmos_coupled_input_lt_idx][
            ..., self.atmos_coupling_var_idx, :, :, :
        ].permute(1, 0, 2, 3, 4, 5)

        return atmos_coupling

    def _make_ocean_coupling(
        self, x: torch.Tensor, coords: CoordSystem
    ) -> torch.Tensor:
        """Make ocean coupling tensor from atmos outputs
        Assumes `x` is a tensor of shape (batch, face, lead_time, variable, height, width).
        Also assumes `x` has been properly normalized.

        Parameters
        ----------
        x : torch.Tensor
            Input data
        coords : CoordSystem
            Input coordinates

        Returns
        -------
        torch.Tensor
            Ocean coupling tensor: shape (lead_time, batch, variable, face, height, width)
        """

        # Subselect the ocean coupling variables -- we use all atmos lead times here so we can average over them
        ocean_coupling = x[:, :, :, self.ocean_coupling_var_idx, :, :]

        # Slice along the lead_time dim, average, and stack along the variable dim
        # (different time-averaged quantities are defined as different variables)
        slices = ocean_coupling.chunk(len(self.ocean_input_times), dim=2)
        ocean_coupling = torch.concat(
            [s.mean(dim=2, keepdim=True) for s in slices], dim=3
        )
        ocean_coupling = ocean_coupling.permute(2, 0, 3, 1, 4, 5)
        return ocean_coupling

    def _normalize_input(self, x: torch.Tensor) -> torch.Tensor:
        """Normalize input data"""

        return (x - self.center) / self.scale

    def _denormalize_output(self, x: torch.Tensor) -> torch.Tensor:
        """Denormalize output data"""

        return x * self.scale + self.center

    def retrieve_valid_ocean_outputs(
        self, x: torch.Tensor, coords: CoordSystem
    ) -> tuple[torch.Tensor, CoordSystem]:
        """Retrieve the valid ocean model outputs from an output data tensor.
        Because we use a dense grid of output times for the coupled model, some of the output times
        for the ocean model may not be valid because it takes a coarser time-step.
        This function will retrieve the valid outputs and return them in a tensor of shape (batch, time, lead_time, variable, face, height, width).

        Parameters
        ----------
        x : torch.Tensor
            Output data tensor
        coords : CoordSystem
            Input coordinates

        Returns
        -------
        torch.Tensor
            Ocean outputs
        CoordSystem
            Output coordinates
        """

        self._validate_output_coords(coords)

        var_dim = list(coords.keys()).index("variable")
        lead_dim = list(coords.keys()).index("lead_time")
        out_coords = coords.copy()
        out_coords["variable"] = np.array(self.ocean_variables)
        out_coords["lead_time"] = np.array(
            [t for t in coords["lead_time"] if t % self.ocean_output_times[0] == 0]
        )

        ocean_outputs = x.index_select(
            dim=var_dim, index=torch.tensor(self.ocean_var_idx, device=x.device)
        )
        ocean_outputs = ocean_outputs.index_select(
            dim=lead_dim, index=torch.tensor(self.ocean_output_lt_idx, device=x.device)
        )
        return ocean_outputs, out_coords

    def retrieve_valid_atmos_outputs(
        self, x: torch.Tensor, coords: CoordSystem
    ) -> tuple[torch.Tensor, CoordSystem]:
        """Retrieve the valid atmospheric model outputs from an output data tensor.
        This function will retrieve the valid outputs and return them in a tensor of shape (batch, time, lead_time, variable, face, height, width).

        Parameters
        ----------
        x : torch.Tensor
            Output data tensor
        coords : CoordSystem
            Input coordinates

        Returns
        -------
        torch.Tensor
            Atmos outputs
        CoordSystem
            Output coordinates
        """

        self._validate_output_coords(coords)

        var_dim = list(coords.keys()).index("variable")

        out_coords = coords.copy()
        out_coords["variable"] = np.array(self.atmos_variables)

        atmos_outputs = x.index_select(
            dim=var_dim, index=torch.tensor(self.atmos_var_idx, device=x.device)
        )

        return atmos_outputs, out_coords

    def _validate_output_coords(self, coords: CoordSystem) -> None:
        """Validate the coordinates passed to the output subselection methods

        Parameters
        ----------
        coords : CoordSystem
            Output coordinates to be validated

        Raises
        ------
        ValueError
            If the coordinates are invalid (missing or incorrect length lead_time dim)
        """
        if "lead_time" not in coords:
            raise ValueError("Lead time is required in the output coordinates")
        if len(coords["lead_time"]) != len(self.atmos_output_times):
            raise ValueError(
                f"Lead time dimension length mismatch between model and coords: expected {len(self.atmos_output_times)}, got {len(coords['lead_time'])}"
            )

    @torch.inference_mode()
    def _forward(
        self,
        x: torch.Tensor,
        coords: CoordSystem,
    ) -> torch.Tensor:

        x = self._normalize_input(x)

        # Forward pass of atmos model first
        atmos_inputs, ocean_inputs = self.prepare_input_data(x, coords)
        atmos_outputs = self.atmos_model(atmos_inputs)

        # Use atmos outputs as coupling for ocean, run forward pass of ocean model
        ocean_inputs.append(self._make_ocean_coupling(atmos_outputs, coords))
        ocean_outputs = self.ocean_model(ocean_inputs)

        output_data = self.prepare_output_data(atmos_outputs, ocean_outputs, coords)
        output_data = self._denormalize_output(output_data)

        return output_data

    def _next_step_inputs(
        self, x: torch.Tensor, coords: CoordSystem
    ) -> tuple[torch.Tensor, CoordSystem]:
        """Get the inputs for the next step of the prognostic model,
        to be used with the model iterator.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor
        coords : CoordSystem
            Input coordinate system

        Returns
        -------
        tuple[torch.Tensor, CoordSystem]
        """

        next_coords = coords.copy()
        next_coords["lead_time"] = coords["lead_time"][-len(self.full_input_times) :]

        next_x = x[:, :, -len(self.full_input_times) :, ...]

        return next_x, next_coords

    @batch_func()
    def __call__(
        self,
        x: torch.Tensor,
        coords: CoordSystem,
    ) -> tuple[torch.Tensor, CoordSystem]:
        """Runs coupled DLESyM model forward 1 step.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor
        coords : CoordSystem
            Input coordinate system

        Returns
        -------
        tuple[torch.Tensor, CoordSystem]
            Output tensor and coordinate system for the prediction
        """

        output_coords = self.output_coords(coords)

        return self._forward(x, coords), output_coords

    @batch_func()
    def _default_generator(
        self, x: torch.Tensor, coords: CoordSystem
    ) -> Generator[tuple[torch.Tensor, CoordSystem], None, None]:

        coords = coords.copy()

        yield x, coords

        while True:
            # Front hook
            x, coords = self.front_hook(x, coords)

            x = self._forward(x, coords)
            coords = self.output_coords(coords)

            # Rear hook
            x, coords = self.rear_hook(x, coords)

            yield x, coords.copy()

            x, coords = self._next_step_inputs(x, coords)

    def create_iterator(
        self, x: torch.Tensor, coords: CoordSystem
    ) -> Iterator[tuple[torch.Tensor, CoordSystem]]:
        """Creates a iterator which can be used to perform time-integration of the
        prognostic model. Will return the initial condition first (0th step).

        Parameters
        ----------
        x : torch.Tensor
            Input tensor
        coords : CoordSystem
            Input coordinate system


        Yields
        ------
        Iterator[tuple[torch.Tensor, CoordSystem]]
            Iterator that generates time-steps of the prognostic model container the
            output data tensor and coordinate system dictionary.
        """
        yield from self._default_generator(x, coords)


class DLESyMLatLon(DLESyM):
    """DLESyM prognostic model supporting lat/lon input and output coordinates.
    This model still uses the HEALPix grid internally, but the first input is regridded
    from lat/lon and the outputs are regridded back to lat/lon upon returning from the model.
    Regridding is done using the `earth2grid` package. For convenience, we expose
    regridding methods that are accessible as `.to_hpx` and `.to_ll`.

    Note
    ----
    See :class:`DLESyM` for more information about the prognostic model. Due to the internal
    regridding, model hooks applied during iteration will need to operate on the HEALPix grid.

    Example
    -------
    .. highlight:: python
    .. code-block:: python

        pkg = DLESyMLatLon.load_default_package()
        model = DLESyMLatLon.load_model(pkg)

        # x and coords are data defined on appropriate lat/lon grid
        x, coords = fetch_data(...)

        # Run model
        x, coords = model(x, coords)

        # Lat-lon outputs
        atmos_outputs, atmos_coords = model.retrieve_valid_atmos_outputs(x, coords)
        ocean_outputs, ocean_coords = model.retrieve_valid_ocean_outputs(x, coords)

        # HEALPix outputs
        atmos_outputs_hpx, atmos_coords_hpx = model.to_hpx(atmos_outputs), model.coords_to_hpx(atmos_coords)
        ocean_outputs_hpx, ocean_coords_hpx = model.to_hpx(ocean_outputs), model.coords_to_hpx(ocean_coords)

    Args
    ----
    *args
        Arguments for :class:`DLESyM`
    **kwargs
        Keyword arguments for :class:`DLESyM`
    """

    @check_extra_imports("dlesym", [OmegaConf, Module, insolation, earth2grid])
    def __init__(
        self,
        atmos_model: torch.nn.Module,
        ocean_model: torch.nn.Module,
        hpx_lat: np.ndarray,
        hpx_lon: np.ndarray,
        nside: int,
        center: np.ndarray,
        scale: np.ndarray,
        atmos_constants: np.ndarray,
        ocean_constants: np.ndarray,
        atmos_input_times: np.ndarray,
        ocean_input_times: np.ndarray,
        atmos_output_times: np.ndarray,
        ocean_output_times: np.ndarray,
        atmos_variables: list[str],
        ocean_variables: list[str],
        atmos_coupling_variables: list[str],
        ocean_coupling_variables: list[str],
    ):

        self.lat = np.linspace(90, -90, 721, endpoint=True)
        self.lon = np.linspace(0, 360, 1440, endpoint=False)

        super().__init__(
            atmos_model=atmos_model,
            ocean_model=ocean_model,
            hpx_lat=hpx_lat,
            hpx_lon=hpx_lon,
            nside=nside,
            center=center,
            scale=scale,
            atmos_constants=atmos_constants,
            ocean_constants=ocean_constants,
            atmos_input_times=atmos_input_times,
            ocean_input_times=ocean_input_times,
            atmos_output_times=atmos_output_times,
            ocean_output_times=ocean_output_times,
            atmos_variables=atmos_variables,
            ocean_variables=ocean_variables,
            atmos_coupling_variables=atmos_coupling_variables,
            ocean_coupling_variables=ocean_coupling_variables,
        )

        self.hpx_grid = earth2grid.healpix.Grid(
            level=int(np.log2(self.nside)),
            pixel_order=earth2grid.healpix.HEALPIX_PAD_XY,
        )
        self.ll_grid = earth2grid.latlon.equiangular_lat_lon_grid(721, 1440)
        self.regrid_to_hpx = earth2grid.get_regridder(self.ll_grid, self.hpx_grid).to(
            torch.float32
        )
        self.regrid_to_ll = earth2grid.get_regridder(self.hpx_grid, self.ll_grid).to(
            torch.float32
        )

    def input_coords(self) -> CoordSystem:
        """Input coordinate system of prognostic model

        Returns
        -------
        CoordSystem
            Coordinate system dictionary
        """
        coords = super().input_coords()
        coords = self.coords_to_ll(coords)

        # Modify to use the base variables instead of the derived variables
        input_variables = [
            v for v in list(coords["variable"]) if v not in ["tau300-700", "ws10m"]
        ]
        input_variables.extend(["u10m", "v10m", "z300", "z700"])
        coords["variable"] = np.array(input_variables)
        return coords

    @batch_coords()
    def output_coords(self, input_coords: CoordSystem) -> CoordSystem:
        """Output coordinate system of the prognostic model

        Parameters
        ----------
        input_coords : CoordSystem
            Input coordinate system

        Returns
        -------
        CoordSystem
            Output coordinate system
        """
        coords = super().output_coords(input_coords)
        coords = self.coords_to_ll(coords)
        return coords

    def to_hpx(self, x: torch.Tensor) -> torch.Tensor:
        """Regrid input data to HEALPix grid. Last 2 dimensions are assumed to be
        (lat, lon)

        Parameters
        ----------
        x : torch.Tensor
            Input tensor

        Returns
        -------
        torch.Tensor
            Regridded tensor, shape (..., 12, nside, nside)
        """
        x = x.to(torch.float32)
        leading_dims = x.shape[:-2]
        x = x.reshape(-1, *x.shape[-2:])
        x = self.regrid_to_hpx(x).reshape(*leading_dims, -1)
        x = x.reshape(*leading_dims, 12, self.nside, self.nside)
        return x

    def to_ll(self, x: torch.Tensor) -> torch.Tensor:
        """Regrid output data from HEALPix grid. Last 3 dimensions are assumed to be
        (face, height, width)

        Parameters
        ----------
        x : torch.Tensor
            Input tensor

        Returns
        -------
        torch.Tensor
            Regridded tensor, shape (..., 721, 1440)
        """
        x = x.to(torch.float32)
        leading_dims = x.shape[:-3]
        x = x.reshape(-1, np.prod(x.shape[-3:]))
        x = self.regrid_to_ll(x).reshape(*leading_dims, len(self.lat), len(self.lon))
        return x

    def coords_to_hpx(self, coords: CoordSystem) -> CoordSystem:
        """Convenience method to pop out lat/lon dimensions from coords and replace with HEALPix"""
        hpx_coords = coords.copy()
        hpx_coords.pop("lat")
        hpx_coords.pop("lon")
        hpx_coords.update(
            {
                "face": np.arange(12),
                "height": np.arange(self.nside),
                "width": np.arange(self.nside),
            }
        )
        for dim in ["face", "height", "width"]:
            hpx_coords.move_to_end(dim)
        return hpx_coords

    def coords_to_ll(self, coords: CoordSystem) -> CoordSystem:
        """Convenience method to pop out HEALPix dimensions from coords and replace with lat/lon"""
        ll_coords = coords.copy()
        ll_coords.pop("face")
        ll_coords.pop("height")
        ll_coords.pop("width")
        ll_coords.update({"lat": self.lat, "lon": self.lon})
        for dim in ["lat", "lon"]:
            ll_coords.move_to_end(dim)
        return ll_coords

    def _nan_interpolate_sst(
        self, sst: torch.Tensor, coords: CoordSystem
    ) -> torch.Tensor:
        """Custom interpolation to fill NaNs over landmasses in SST data."""

        da_sst = xr.DataArray(sst.cpu().numpy(), dims=coords.keys())
        da_interp = da_sst.interpolate_na(
            dim="lon", method="linear", use_coordinate=False
        )

        # Second pass: roll, interpolate along longitude, and unroll
        roll_amount_lon = int(len(da_interp.lon) / 2)
        da_double_interp = (
            da_interp.roll(lon=roll_amount_lon, roll_coords=False)
            .interpolate_na(dim="lon", method="linear", use_coordinate=False)
            .roll(lon=len(da_interp.lon) - roll_amount_lon, roll_coords=False)
        )

        # Third pass do a similar roll along latitude
        roll_amount_lat = int(len(da_double_interp.lat) / 2)
        da_triple_interp = (
            da_double_interp.roll(lat=roll_amount_lat, roll_coords=False)
            .interpolate_na(dim="lat", method="linear", use_coordinate=False)
            .roll(lat=len(da_double_interp.lat) - roll_amount_lat, roll_coords=False)
        )

        return torch.from_numpy(da_triple_interp.values).to(sst.device)

    def _prepare_derived_variables(
        self, x: torch.Tensor, coords: CoordSystem
    ) -> tuple[torch.Tensor, CoordSystem]:
        """Prepare derived variables for the DLESyM model.

        This method handles the preparation of derived variables from the input tensor
        and coordinates. It ensures that the derived variables are correctly computed,
        and performs NaN-interpolation on the SST data.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor
        coords : CoordSystem
            Input coordinate system

        Returns
        -------
        tuple[torch.Tensor, CoordSystem]
            Output tensor and coordinate system for the derived variables
        """

        prep_coords = coords.copy()

        # Fetch the base variables
        base_vars = list(prep_coords["variable"])
        src_vars = {
            v: x[..., base_vars.index(v) : base_vars.index(v) + 1, :, :]
            for v in base_vars
        }

        # Compute the derived variables
        out_vars = {
            "ws10m": torch.sqrt(src_vars["u10m"] ** 2 + src_vars["v10m"] ** 2),
            "tau300-700": src_vars["z300"] - src_vars["z700"],
        }
        out_vars.update(src_vars)

        # Fill SST nans by custom interpolation
        out_vars["sst"] = self._nan_interpolate_sst(out_vars["sst"], coords)

        # Update the tensor with the derived variables and return
        prep_coords["variable"] = np.array(self.atmos_variables + self.ocean_variables)
        x_out = torch.empty(
            *[v.shape[0] for v in prep_coords.values()], device=x.device
        )
        for i, v in enumerate(prep_coords["variable"]):
            x_out[..., i : i + 1, :, :] = out_vars[v]

        return x_out, prep_coords

    @batch_func()
    def __call__(
        self, x: torch.Tensor, coords: CoordSystem
    ) -> tuple[torch.Tensor, CoordSystem]:
        """Runs coupled DLESyM model forward 1 step, regridding to/from HEALPix grid

        Parameters
        ----------
        x : torch.Tensor
            Input tensor
        coords : CoordSystem
            Input coordinate system

        Returns
        -------
        tuple[torch.Tensor, CoordSystem]
            Output tensor and coordinate system for the prediction
        """
        output_coords = self.output_coords(coords)

        x, coords = self._prepare_derived_variables(x, coords)

        x = self.to_hpx(x)
        x = self._forward(x, self.coords_to_hpx(coords))
        x = self.to_ll(x)
        return x, output_coords

    @batch_func()
    def _default_generator(
        self, x: torch.Tensor, coords: CoordSystem
    ) -> Generator[tuple[torch.Tensor, CoordSystem], None, None]:

        coords = coords.copy()

        base_vars = coords["variable"]

        x, coords = self._prepare_derived_variables(x, coords)

        yield x, coords

        x = self.to_hpx(x)

        while True:
            # Front hook
            x, coords = self.front_hook(x, coords)

            x = self._forward(x, self.coords_to_hpx(coords))

            # Output coords expects the input variable set to include base variables,
            # but will return the ouptut variables with the derived variables
            base_coords = coords.copy()
            base_coords["variable"] = base_vars
            coords = self.output_coords(base_coords)

            # Rear hook
            x, coords = self.rear_hook(x, coords)

            yield self.to_ll(x), coords.copy()

            x, coords = self._next_step_inputs(x, coords)



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .arco import ARCOLexicon
from .cds import CDSLexicon
from .gefs import GEFSLexicon, GEFSLexiconSel
from .gfs import GFSLexicon
from .hrrr import HRRRFXLexicon, HRRRLexicon
from .ifs import IFSLexicon
from .imerg import IMERGLexicon
from .ncar import NCAR_ERA5Lexicon
from .wb2 import WB2ClimatetologyLexicon, WB2Lexicon



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

from earth2studio.models.dx.base import DiagnosticModel
from earth2studio.models.dx.climatenet import ClimateNet  # noqa
from earth2studio.models.dx.corrdiff import CorrDiffTaiwan  # noqa
from earth2studio.models.dx.derived import (
    DerivedRH,
    DerivedRHDewpoint,
    DerivedVPD,
    DerivedWS,
)
from earth2studio.models.dx.identity import Identity  # noqa
from earth2studio.models.dx.precipitation_afno import PrecipitationAFNO  # noqa
from earth2studio.models.dx.precipitation_afno_v2 import PrecipitationAFNOv2  # noqa
from earth2studio.models.dx.tc_tracking import (
    TCTrackerVitart,
    TCTrackerWuDuan,
)
from earth2studio.models.dx.wind_gust import WindgustAFNO  # noqa

# TODO: Remove upon physics-nemo update...
# package turned on logging of warnings in 1.1.0, this is silencing them
warnings.filterwarnings("ignore")
warnings.filterwarnings("ignore", category=SyntaxWarning)
warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=FutureWarning)



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import TypeVar

try:
    import onnxruntime as ort
    from onnxruntime import InferenceSession
except ImportError:
    ort = None
    InferenceSession = TypeVar("InferenceSession")  # type: ignore
import torch


def create_ort_session(
    onnx_file: str,
    device: torch.device = torch.device("cpu", 0),
) -> InferenceSession:
    """Create ORT session on specified device

    Parameters
    ----------
    onnx_file : str
        ONNX file
    device : torch.device, optional
        Device for session to run on, by default "cpu"

    Returns
    -------
    ort.InferenceSession
        ORT inference session
    """
    if ort is None:
        raise ImportError(
            "onnxruntime (onnxruntime-gpu) is required for this model. See model install notes for details.\n"
            + "https://nvidia.github.io/earth2studio/userguide/about/install.html#model-dependencies"
        )
    options = ort.SessionOptions()
    options.enable_cpu_mem_arena = False
    options.enable_mem_pattern = False
    options.enable_mem_reuse = False
    options.intra_op_num_threads = 1
    options.log_severity_level = 3

    # That will trigger a FileNotFoundError
    os.stat(onnx_file)
    if device.type == "cuda":
        if device.index is None:
            device_index = torch.cuda.current_device()
        else:
            device_index = device.index

        providers = [
            (
                "CUDAExecutionProvider",
                {
                    "device_id": device_index,
                },
            ),
            "CPUExecutionProvider",
        ]
    else:
        providers = [
            "CPUExecutionProvider",
        ]

    ort_session = ort.InferenceSession(
        onnx_file,
        sess_options=options,
        providers=providers,
    )

    return ort_session



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
from collections.abc import Generator, Iterator
from datetime import datetime, timezone

import numpy as np
import torch

try:
    from aurora import Aurora as Aurora_model
    from aurora import Batch, Metadata
except ImportError:
    Aurora_model = None
    Batch = None
    Metadata = None

from earth2studio.models.auto import AutoModelMixin, Package
from earth2studio.models.batch import batch_coords, batch_func
from earth2studio.models.px.base import PrognosticModel
from earth2studio.models.px.utils import PrognosticMixin
from earth2studio.utils import check_extra_imports, handshake_coords, handshake_dim
from earth2studio.utils.type import CoordSystem

VARIABLES = [
    "z1000",
    "z925",
    "z850",
    "z700",
    "z600",
    "z500",
    "z400",
    "z300",
    "z250",
    "z200",
    "z150",
    "z100",
    "z50",
    "q1000",
    "q925",
    "q850",
    "q700",
    "q600",
    "q500",
    "q400",
    "q300",
    "q250",
    "q200",
    "q150",
    "q100",
    "q50",
    "t1000",
    "t925",
    "t850",
    "t700",
    "t600",
    "t500",
    "t400",
    "t300",
    "t250",
    "t200",
    "t150",
    "t100",
    "t50",
    "u1000",
    "u925",
    "u850",
    "u700",
    "u600",
    "u500",
    "u400",
    "u300",
    "u250",
    "u200",
    "u150",
    "u100",
    "u50",
    "v1000",
    "v925",
    "v850",
    "v700",
    "v600",
    "v500",
    "v400",
    "v300",
    "v250",
    "v200",
    "v150",
    "v100",
    "v50",
    "msl",
    "u10m",
    "v10m",
    "t2m",
]

ATMOS_LEVELS = [1000, 925, 850, 700, 600, 500, 400, 300, 250, 200, 150, 100, 50]


# Adapted from https://microsoft.github.io/aurora/example_era5.html
@check_extra_imports("aurora", [Aurora_model, Batch, Metadata])
class Aurora(torch.nn.Module, AutoModelMixin, PrognosticMixin):
    """Aurora 0.25 degree global forecast model. This model consists of single
    auto-regressive model with a time-step size of 6 hours. This model operates on
    0.25 degree lat-lon grid (720, 1440) equirectangular grid with 4 surface-level
    variables, 5 atmospheric variables with 13 pressure levels and 3 static variables.

    Note
    ----
    This model uses the checkpoints from the original publication.
    For additional information see the following resources:

    - https://arxiv.org/abs/2405.13063
    - https://github.com/microsoft/aurora
    - https://huggingface.co/microsoft/aurora

    Warning
    -------
    We encourage users to familiarize themselves with the license restrictions of this
    model's checkpoints.

    Parameters
    ----------
    core_model : torch.nn.Module
        Core Aurora model
    z : torch.Tensor
        Geopotential
    slt : torch.Tensor
        Soil type
    lsm : torch.Tensor
        Land sea mask
    """

    def __init__(
        self,
        core_model: torch.nn.Module,
        z: torch.Tensor,
        slt: torch.Tensor,
        lsm: torch.Tensor,
    ) -> None:
        super().__init__()

        self.model = core_model
        self.register_buffer("z", z)
        self.register_buffer("slt", slt)
        self.register_buffer("lsm", lsm)

        self._input_coords = OrderedDict(
            {
                "batch": np.empty(0),
                "time": np.empty(0),
                "lead_time": np.array(
                    [np.timedelta64(-6, "h"), np.timedelta64(0, "h")]
                ),
                "variable": np.array(VARIABLES),
                "lat": np.linspace(90, -90, 720, endpoint=False),
                "lon": np.linspace(0, 360, 1440, endpoint=False),
            }
        )

        self._output_coords = OrderedDict(
            {
                "batch": np.empty(0),
                "time": np.empty(0),
                "lead_time": np.array([np.timedelta64(6, "h")]),
                "variable": np.array(VARIABLES),
                "lat": np.linspace(90, -90, 720, endpoint=False),
                "lon": np.linspace(0, 360, 1440, endpoint=False),
            }
        )
        self.device = torch.ones(1).device  # Hack to get default device
        self.preds_idx = 0

    def input_coords(self) -> CoordSystem:
        """Input coordinate system of the prognostic model

        Returns
        -------
        CoordSystem
            Coordinate system dictionary
        """
        return self._input_coords.copy()

    @batch_coords()
    def output_coords(
        self,
        input_coords: CoordSystem,
    ) -> CoordSystem:
        """Output coordinate system of the prognostic model

        Parameters
        ----------
        input_coords : CoordSystem
            Input coordinate system to transform into output_coords

        Returns
        -------
        CoordSystem
            Coordinate system dictionary
        """
        output_coords = self._output_coords.copy()

        test_coords = input_coords.copy()
        test_coords["lead_time"] = (
            test_coords["lead_time"] - input_coords["lead_time"][-1]
        )
        target_input_coords = self.input_coords()
        for i, key in enumerate(target_input_coords):
            if key not in ["batch", "time"]:
                handshake_dim(test_coords, key, i)
                handshake_coords(test_coords, target_input_coords, key)

        output_coords["batch"] = input_coords["batch"]
        output_coords["time"] = input_coords["time"]

        output_coords["lead_time"] = (
            input_coords["lead_time"][-1] + output_coords["lead_time"]
        )

        return output_coords

    @classmethod
    def load_default_package(cls) -> Package:
        """Load prognostic package"""
        return Package(
            "hf://microsoft/aurora@refs%2Fpr%2F1",
            cache_options={
                "cache_storage": Package.default_cache("aurora"),
                "same_names": True,
            },
        )

    @classmethod
    @check_extra_imports("aurora", [Aurora_model, Batch, Metadata])
    def load_model(
        cls,
        package: Package,
    ) -> PrognosticModel:
        """Load prognostic from package"""

        # Import the static variables: z, slt, lsm
        z = torch.from_numpy(
            np.load(package.resolve("static-npy/aurora-0.25-static-z.npy"))[:-1]
        )
        slt = torch.from_numpy(
            np.load(package.resolve("static-npy/aurora-0.25-static-slt.npy"))[:-1]
        )
        lsm = torch.from_numpy(
            np.load(package.resolve("static-npy/aurora-0.25-static-lsm.npy"))[:-1]
        )

        # Load 0.25 degrees resolution Aurora pretrained model
        aurora_model = package.resolve("aurora-0.25-pretrained.ckpt")
        # The pretrained version does not use LoRA.
        model = Aurora_model(use_lora=False)
        model.load_checkpoint_local(aurora_model)
        model.eval()

        return cls(model, z, slt, lsm)

    def _prepare_input(self, input: torch.Tensor, coords: CoordSystem) -> Batch:
        """Prepares input Batch"""
        len_atmos_levels = len(ATMOS_LEVELS)
        # Only len 1 time array is supported by Auroral model
        ts = (
            (coords["time"][0] + coords["lead_time"][-1]).astype("datetime64[s]")
            - np.datetime64("1970-01-01T00:00:00Z")
        ) / np.timedelta64(1, "s")
        time = datetime.fromtimestamp(ts, tz=timezone.utc)

        # input variable order: z, q, t, u, v, msl, u10m, v10m, t2m
        batch = Batch(
            surf_vars={
                # select time points `i` and `i - 1`
                "2t": input[:, 0, -2:, len_atmos_levels * 5 + 3],
                "10u": input[:, 0, -2:, len_atmos_levels * 5 + 1],
                "10v": input[:, 0, -2:, len_atmos_levels * 5 + 2],
                "msl": input[:, 0, -2:, len_atmos_levels * 5],
            },
            static_vars={
                "z": self.z,
                "slt": self.slt,
                "lsm": self.lsm,
            },
            atmos_vars={
                "t": input[:, 0, -2:, len_atmos_levels * 2 : len_atmos_levels * 3],
                "u": input[:, 0, -2:, len_atmos_levels * 3 : len_atmos_levels * 4],
                "v": input[:, 0, -2:, len_atmos_levels * 4 : len_atmos_levels * 5],
                "q": input[:, 0, -2:, len_atmos_levels : len_atmos_levels * 2],
                "z": input[:, 0, -2:, :len_atmos_levels],
            },
            metadata=Metadata(
                lat=torch.from_numpy(coords["lat"]),
                lon=torch.from_numpy(coords["lon"]),
                time=(time,),
                atmos_levels=tuple(int(level) for level in ATMOS_LEVELS),
                rollout_step=self.preds_idx,
            ),
        )

        return batch

    def _prepare_output(self, output: Batch, coords: CoordSystem) -> torch.Tensor:
        # retrieving tensor output from Batch format
        x = torch.cat(
            [
                output.atmos_vars["z"],
                output.atmos_vars["q"],
                output.atmos_vars["t"],
                output.atmos_vars["u"],
                output.atmos_vars["v"],
                output.surf_vars["msl"].unsqueeze(2),
                output.surf_vars["10u"].unsqueeze(2),
                output.surf_vars["10v"].unsqueeze(2),
                output.surf_vars["2t"].unsqueeze(2),
            ],
            dim=2,
        )
        x = x.view(-1, 1, *x.shape[1:])
        # According to https://github.com/microsoft/aurora/blob/main/aurora/rollout.py, the static variables of next step are from the previous step.
        self.z = output.static_vars["z"]
        self.slt = output.static_vars["slt"]
        self.lsm = output.static_vars["lsm"]

        return x

    @torch.inference_mode()
    def _forward(
        self,
        x: torch.Tensor,
        coords: CoordSystem,
    ) -> torch.Tensor:

        # output shape: [b,t,1,69,720,1440], batch, time, lead_time, variables, lat, lon
        out = torch.empty_like(x[:, :, :1])
        # The aurora model can only process one time per iteration, so loop if multiple
        for t in range(coords["time"].shape[0]):
            batch_coords = coords.copy()
            batch_coords["time"] = batch_coords["time"][t : t + 1]
            # x shape: [b,1,2,69,720,1440], batch, lead_time, variables, lat, lon
            input_batch = self._prepare_input(x[:, t : t + 1], batch_coords)
            # Convert tensor to Batch, atmos_vars is the first 65 variables, surf_vars is the last 4 variables of x
            output_batch = self.model(input_batch)
            # Convert Batch to tensor
            out[:, t : t + 1] = self._prepare_output(output_batch, coords)

        return out

    @batch_func()
    def __call__(
        self,
        x: torch.Tensor,
        coords: CoordSystem,
    ) -> tuple[torch.Tensor, CoordSystem]:
        """Runs prognostic model 1 step.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor
        coords : CoordSystem
            Input coordinate system

        Returns
        -------
        tuple[torch.Tensor, CoordSystem]
            Output tensor and coordinate system 6 hours in the future
        """

        output_coords = self.output_coords(coords)

        x = self._forward(x, coords)

        return x, output_coords

    @batch_func()
    def _default_generator(
        self, x: torch.Tensor, coords: CoordSystem
    ) -> Generator[tuple[torch.Tensor, CoordSystem], None, None]:
        coords = coords.copy()

        self.output_coords(coords)

        yield x[:, :, 1:], coords

        while True:
            # Front hook
            x, coords = self.front_hook(x, coords)
            init_x = x[:, :, 1:].clone()

            # Forward pass
            x = self._forward(x, coords)

            self.preds_idx = self.preds_idx + 1

            coords["lead_time"] = (
                coords["lead_time"]
                + self.output_coords(self.input_coords())["lead_time"]
            )
            # Concat the step now and prediction for next step
            x = torch.cat([init_x, x], dim=2)
            x = x.clone()

            # Rear hook for first predicted step
            coords_out = coords.copy()
            coords_out["lead_time"] = coords["lead_time"][-1]
            x[:, :, 1:], coords_out = self.rear_hook(x[:, :, 1:], coords_out)

            yield x[:, :, 1:], coords_out

    def create_iterator(
        self, x: torch.Tensor, coords: CoordSystem
    ) -> Iterator[tuple[torch.Tensor, CoordSystem]]:
        """Creates a iterator which can be used to perform time-integration of the
        prognostic model. Will return the initial condition first (0th step).

        Parameters
        ----------
        x : torch.Tensor
            Input tensor
        coords : CoordSystem
            Input coordinate system


        Yields
        ------
        Iterator[tuple[torch.Tensor, CoordSystem]]
            Iterator that generates time-steps of the prognostic model container the
            output data tensor and coordinate system dictionary.
        """
        yield from self._default_generator(x, coords)



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable

import numpy as np

from .base import LexiconType

LEVELS = [50, 100, 150, 200, 250, 300, 400, 500, 600, 700, 850, 925, 1000]


class WB2Lexicon(metaclass=LexiconType):
    """WeatherBench Lexicon
    WeatherBench specified <Variable ID>::<Pressure Level>

    Note
    ----
    Variable named based on ERA5 names, see WB2 docs for resources.

    - https://weatherbench2.readthedocs.io/en/latest/data-guide.html
    - Dew point temperature at 2m seems to be all NaNs
    """

    VOCAB = {
        "u10m": "10m_u_component_of_wind::",
        "v10m": "10m_v_component_of_wind::",
        "t2m": "2m_temperature::",
        "sp": "surface_pressure::",
        "msl": "mean_sea_level_pressure::",
        "sst": "sea_surface_temperature::",
        "tcwv": "total_column_water_vapour::",
        "tp06": "total_precipitation_6hr::",
        "tp12": "total_precipitation_12hr::",
        "tp24": "total_precipitation_24hr::",
    }
    VOCAB.update({f"u{level}": f"u_component_of_wind::{level}" for level in LEVELS})
    VOCAB.update({f"v{level}": f"v_component_of_wind::{level}" for level in LEVELS})
    VOCAB.update({f"w{level}": f"vertical_velocity::{level}" for level in LEVELS})
    VOCAB.update({f"z{level}": f"geopotential::{level}" for level in LEVELS})
    VOCAB.update({f"t{level}": f"temperature::{level}" for level in LEVELS})
    VOCAB.update({f"q{level}": f"specific_humidity::{level}" for level in LEVELS})
    VOCAB.update({f"r{level}": f"relative_humidity::{level}" for level in LEVELS})

    @classmethod
    def get_item(cls, val: str) -> tuple[str, Callable]:
        """Return name in WeatherBench vocabulary."""
        wb2_key = cls.VOCAB[val]

        if wb2_key.split("::")[0] == "relative_humidity":

            def mod(x: np.array) -> np.array:
                """Relative humidty in WeatherBench uses older calculation and does
                not scale by 100 natively. Not recommended for use, see IFS method for
                more modern calculation.
                https://github.com/google-research/weatherbench2/blob/main/weatherbench2/derived_variables.py#L468
                """
                return x * 100

        else:

            def mod(x: np.array) -> np.array:
                """Modify name (if necessary)."""
                return x

        return wb2_key, mod


class WB2ClimatetologyLexicon(metaclass=LexiconType):
    """WeatherBench Climatology Lexicon
    WeatherBench specified <Variable ID>::<Pressure Level>

    Note
    ----
    Variable named based on ERA5 names, see WB2 docs for resources.

    - https://weatherbench2.readthedocs.io/en/latest/data-guide.html
    - Dew point temperature at 2m seems to be all NaNs
    """

    VOCAB = {
        "u10m": "10m_u_component_of_wind::",
        "v10m": "10m_v_component_of_wind::",
        "t2m": "2m_temperature::",
        "sp": "surface_pressure::",
        "msl": "mean_sea_level_pressure::",
        "sst": "sea_surface_temperature::",
        "tcwv": "total_column_water_vapour::",
        "tp06": "total_precipitation_6hr::",
        "tp06sdf": "total_precipitation_6hr_seeps_dry_fraction::",
        "tp06st": "total_precipitation_6hr_seeps_threshold::",
        "tp12": "total_precipitation_12hr::",
        "tp24": "total_precipitation_24hr::",
        "tp24sdf": "total_precipitation_24hr_seeps_dry_fraction::",
        "tp24st": "total_precipitation_24hr_seeps_threshold::",
    }
    VOCAB.update({f"u{level}": f"u_component_of_wind::{level}" for level in LEVELS})
    VOCAB.update({f"v{level}": f"v_component_of_wind::{level}" for level in LEVELS})
    VOCAB.update({f"w{level}": f"vertical_velocity::{level}" for level in LEVELS})
    VOCAB.update({f"z{level}": f"geopotential::{level}" for level in LEVELS})
    VOCAB.update({f"t{level}": f"temperature::{level}" for level in LEVELS})
    VOCAB.update({f"q{level}": f"specific_humidity::{level}" for level in LEVELS})
    VOCAB.update({f"r{level}": f"relative_humidity::{level}" for level in LEVELS})

    @classmethod
    def get_item(cls, val: str) -> tuple[str, Callable]:
        """Return name in WeatherBench vocabulary."""
        wb2_key = cls.VOCAB[val]

        if wb2_key.split("::")[0] == "relative_humidity":

            def mod(x: np.array) -> np.array:
                """Relative humidty in WeatherBench uses older calculation and does
                not scale by 100 natively. Not recommended for use, see IFS method for
                more modern calculation.
                https://github.com/google-research/weatherbench2/blob/main/weatherbench2/derived_variables.py#L468
                """
                return x * 100

        else:

            def mod(x: np.array) -> np.array:
                """Modify name (if necessary)."""
                return x

        return wb2_key, mod



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from collections.abc import Iterator
from typing import Any, Protocol, runtime_checkable

import torch

from earth2studio.utils.type import CoordSystem


@runtime_checkable
class PrognosticModel(Protocol):
    """Prognostic model interface"""

    def __call__(
        self,
        x: torch.Tensor,
        coords: CoordSystem,
    ) -> tuple[torch.Tensor, CoordSystem]:
        """Forward pass of the prognostic model, time integrating a single time-step

        Parameters
        ----------
        x : torch.Tensor
            Input tensor intended to apply diagnostic function on
        coords : CoordSystem
            Ordered dict representing coordinate system that describes the tensor

        Returns
        -------
        tuple[torch.Tensor, CoordSystem]
            Output tensor and coordinate dictionary one time-step into the future
        """
        pass

    def create_iterator(
        self, x: torch.Tensor, coords: CoordSystem
    ) -> Iterator[tuple[torch.Tensor, CoordSystem]]:
        """Creates a iterator which can be used to perform time-integration of the
        prognostic model. Will return the initial condition first (0th step).

        Parameters
        ----------
        x : torch.Tensor
            Input tensor, which can be viewed as the initial state of the prognositc
        coords : CoordSystem
            Input coordinate system

        Yields
        ------
        Iterator[tuple[torch.Tensor, CoordSystem]]
            Iterator that generates time-steps of the prognostic model container the
            output data tensor and coordinate system dictionary.
        """
        pass

    def input_coords(self) -> CoordSystem:
        """Input coordinate system of prognostic model, time dimension should contain
        time-delta objects

        Returns
        -------
        CoordSystem
            Coordinate system dictionary
        """
        pass

    def output_coords(self, input_coords: CoordSystem) -> CoordSystem:
        """Output coordinate system of the prognostic model give an input coordinate
        system.

        Parameters
        ----------
        input_coords : CoordSystem
            Input coordinate system to transform into output_coords

        Returns
        -------
        CoordSystem
            Coordinate system dictionary

        Raises
        ------
        ValueError
            If input_coords are not valid
        """
        pass

    def to(self, device: Any) -> PrognosticModel:
        """Moves prognostic model onto inference device, this is typically satisfied via
        `torch.nn.Module`.

        Parameters
        ----------
        device : Any
            Object representing the inference device, typically `torch.device` or str

        Returns
        -------
        PrognosticModel
            Returns instance of prognostic
        """
        pass



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable
from functools import wraps
from importlib.util import find_spec
from typing import Any, TypeVar, cast

F = TypeVar("F", bound=Callable[..., Any])
C = TypeVar("C", bound=type[Any])


class ExtraDependencyError(ImportError):
    """Custom error for missing extra dependencies."""

    def __init__(self, extra_name: str, obj_name: str):
        super().__init__(
            f"Extra dependency group '{extra_name}' is required for {obj_name}.\n"
            f"Install with: uv pip install earth2studio --extra {extra_name} \n"
            f"or refer to the install documentation."
        )


def check_extra_imports(
    extra_name: str, package_obj: Any | list[Any]
) -> Callable[[Any], Any]:
    """Decorator to handle validating optional dependencies are installed for both
    functions and classes.

    Parameters
    ----------
    extra_name : str
        Name of the extra/optional dependency group present in the pyproject.toml
        E.g. 'corrdiff', 'fcn'
    package_obj : Any | list[Any]
        Object(s) that should be instantiated / imported. If it is a string, it will be
        treated as a module path and checked if it is available. Otherwise this will
        check if the object is instantiated.
    """

    def _check_dependencies(obj_name: Any) -> None:
        package_objs = package_obj if isinstance(package_obj, list) else [package_obj]
        for pkg in package_objs:
            # If package is a string (module path like module.submodule), check if the
            #  module is available.
            if isinstance(pkg, str) and find_spec(pkg) is None:
                raise ExtraDependencyError(extra_name, obj_name)
            elif pkg is None:
                raise ExtraDependencyError(extra_name, obj_name)

    def _decorator(obj: F | C) -> F | C:
        if isinstance(obj, type):
            original_init = obj.__init__  # type: ignore

            @wraps(original_init)
            def _wrapped_init(self: Any, *args: Any, **kwargs: Any) -> None:
                _check_dependencies(obj.__name__)
                original_init(self, *args, **kwargs)

            obj.__init__ = _wrapped_init  # type: ignore
            return cast(C, obj)
        else:
            # Handle function decoration
            @wraps(obj)
            def _wrapper(*args: Any, **kwargs: Any) -> Any:
                _check_dependencies(f"{obj.__module__}.{obj.__name__}")
                return obj(*args, **kwargs)

            return cast(F, _wrapper)

    return _decorator



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable

import numpy as np

from .base import LexiconType


class IFSLexicon(metaclass=LexiconType):
    """Integrated Forecast System Lexicon
    IFS specified <Grib Parameter ID>::<Level Type>::<Level>

    Note
    ----
    Additional resources:
    https://codes.ecmwf.int/grib/param-db/?filter=grib2
    https://www.ecmwf.int/en/forecasts/datasets/open-data
    Best bet is to download an index file from the AWS bucket and read it
    """

    @staticmethod
    def build_vocab() -> dict[str, str]:
        """Create HRRR vocab dictionary"""
        sfc_variables = {
            "u10m": "10u::sfc::",
            "v10m": "10v::sfc::",
            "u100m": "100u::sfc::",
            "v100m": "100v::sfc::",
            "t2m": "2t::sfc::",
            "d2m": "2d::sfc::",
            "sp": "sp::sfc::",
            "msl": "msl::sfc::",
            "tcw": "tcw::sfc::",  #
            "tcwv": "tcwv::sfc::",
            "tp": "tp::sfc::",
            "skt": "skt::sfc::",
            "slor": "slor::sfc::",
            "sdor": "sdor::sfc::",
            "lsm": "lsm::sfc::",
            "zsl": "z::sfc::",
        }
        soil_variables = {
            "swvl1": "vsw::sl::1",
            "swvl2": "vsw::sl::2",
            "stl1": "sot::sl::1",
            "stl2": "sot::sl::2",
        }
        prs_levels = [
            50,
            100,
            150,
            200,
            250,
            300,
            400,
            500,
            600,
            700,
            850,
            925,
            1000,
        ]
        prs_names = ["u", "v", "w", "gh", "t", "r", "q"]
        e2s_id = ["u", "v", "w", "z", "t", "r", "q"]
        prs_variables = {}
        for id, variable in zip(e2s_id, prs_names):
            for level in prs_levels:
                prs_variables[f"{id}{level:d}"] = f"{variable}::pl::{level}"

        return {**sfc_variables, **soil_variables, **prs_variables}

    VOCAB = build_vocab()

    @classmethod
    def get_item(cls, val: str) -> tuple[str, Callable]:
        """Retrieve name from IFS vocabulary."""
        ifs_key = cls.VOCAB[val]
        if ifs_key.split("::")[0] == "gh":

            def mod(x: np.array) -> np.array:
                """Modify data value (if necessary)."""
                return x * 9.81

        else:

            def mod(x: np.array) -> np.array:
                """Modify data value (if necessary)."""
                return x

        return ifs_key, mod



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
from collections.abc import Generator

import numpy as np
import torch
from loguru import logger

from earth2studio.data import DataSource, fetch_data
from earth2studio.models.px import PrognosticModel
from earth2studio.perturbation.base import Perturbation
from earth2studio.utils import handshake_dim, handshake_size
from earth2studio.utils.time import to_time_array
from earth2studio.utils.type import CoordSystem, TimeArray


class HemisphericCentredBredVector:
    """Bred Vector perturbation method, following the approach introduced in
    'Huge Ensembles Part I: Design of Ensemble Weather Forecasts using Spherical Fourier
    Neural Operators'. The vector is bred by advancing in time. The bred vector is
    scaled seperately in the northern and southern extra-tropics and interpolated in the
    tropics. Additionally, it applies a centred perturbation, ie generating two
    perturbed stated by adding and subtracting the bred vector, respecitvely.


    Parameters
    ----------
    model : PrognosticModel
        Dynamical model, typically this is the prognostic AI model
    data : DataSource
        data source for obtaining warmup time steps
    seeding_perturbation_method : Perturbation, optional
        Method to seed the Bred Vector perturbation
    noise_amplitude : float | Tensor, optional
        Noise amplitude, by default 0.05. If a tensor, this must be broadcastable with
        the input data
    integration_steps : int, optional
        Number of integration steps to use in forward call, by default 20

    Note
    ----
    For additional information:

    - https://arxiv.org/abs/2408.03100
    - https://journals.ametsoc.org/view/journals/bams/74/12/1520-0477_1993_074_2317_efantg_2_0_co_2.xml
    - https://en.wikipedia.org/wiki/Bred_vector
    """

    def __init__(
        self,
        model: PrognosticModel,
        data: DataSource,
        seeding_perturbation_method: Perturbation,
        noise_amplitude: float | torch.Tensor = 0.35,
        integration_steps: int = 3,
    ) -> None:
        self.model = model
        self.data = data
        self.noise_amplitude = (
            noise_amplitude
            if isinstance(noise_amplitude, torch.Tensor)
            else torch.Tensor(
                [noise_amplitude] * len(self.model.input_coords()["variable"])
            )[:, None, None]
        )
        self.integration_steps = integration_steps
        self.seeding_perturbation_method = seeding_perturbation_method
        self.set_clip_indices()
        self._residual: list[torch.Tensor] = []

    @torch.inference_mode()
    def create_generator(
        self, time: TimeArray, generator_size: int = 1, device: torch.device = "cpu"
    ) -> Generator[torch.Tensor, None, None]:
        """Creates and initializes the perturbation generator"""
        # Initialize your IC or other necessary components
        batch_size = generator_size // 2
        input_coords = self.model.input_coords()

        time = to_time_array(time)
        warmup_times = (
            time
            + np.arange(-self.integration_steps, 1)
            * self.model.output_coords(input_coords)["lead_time"]
        )
        input_data, data_coords = fetch_data(
            source=self.data,
            time=warmup_times,
            variable=input_coords["variable"],
            lead_time=input_coords["lead_time"],
            device="cpu",
        )
        coords = OrderedDict(
            [("batch", np.arange(batch_size))] + list(data_coords.items())
        )

        if input_coords["lead_time"].shape[0] > 1:
            logger.warning(
                "Input data / models that require multiple lead times may lead to unexpected behavior"
            )

        # get unperturbed intital state, assuming tensor always has 6 dims
        xunp = input_data[:1].repeat(batch_size, 1, 1, 1, 1, 1).to(device)
        # Commenting here, not tested for when multiple lead times are needed
        # May work...
        coords["time"] = data_coords["time"][:1]

        # generate perturbed initial state
        xper, coords = self.seeding_perturbation_method(xunp, coords)

        for ii in range(self.integration_steps):
            xunp, _ = self.model(xunp, coords)
            xper, _ = self.model(xper, coords)

            dx = xper - xunp
            hem_norm = self.hemispheric_norm(dx, device)

            # if zero elements are requierd, replace NaNs in scaled dx with 0
            if (hem_norm == 0).any():
                raise ValueError(
                    "zero element in hemispheric norm, maybe noise amplification too small?"
                )

            # scale dx
            dx = self.noise_amplitude * (dx / hem_norm)

            xunp = (
                input_data[ii + 1]
                .unsqueeze(dim=0)
                .repeat(batch_size, 1, 1, 1, 1, 1)
                .to(device)
            )
            coords["time"] = data_coords["time"][ii + 1 : ii + 2]
            xper = xunp + dx
            # self.force_non_neg(xper[i : i + 1])

        # Yield single batches in alternating order to keep perturbation centered
        for i in range(batch_size):
            yield self.force_non_neg(xper[i : i + 1])
            yield self.force_non_neg(xunp[i : i + 1] - dx[i : i + 1])

    def set_clip_indices(self) -> None:
        """If humidity and tcwv in variable set, add to list of variables to clip"""
        self.clip_idcs = []
        for ii, var in enumerate(self.model.input_coords()["variable"]):
            if var[0] == "q" or var == "tcwv" or var[0] == "r" or var[:2] == "tp":
                self.clip_idcs.append(ii)
        return

    def hemispheric_norm(self, x: torch.Tensor, device: torch.device) -> torch.Tensor:
        """Calculate norm of fields speperately in northern and southern extra-tropics

        Parameters
        ----------
        x : torch.Tensor
            Input tensor to apply norm to
        """
        nvar, nlat, nlon = x.shape[-3:]
        ex_tropic = int(
            70 / 180 * nlat
        )  # using 70deg for extra-tropics like in hens publication
        in_tropic = nlat - 2 * ex_tropic

        jac = torch.sin(torch.linspace(0, torch.pi, nlat))[:, None].to(device)
        weights = (torch.pi / nlat) * (2 * torch.pi / nlon) * jac
        weights = weights / weights.mean()

        # compute norms on northern and southern hemisphere
        north = torch.sqrt(
            torch.mean(
                x[..., :ex_tropic, :] ** 2 * weights[:ex_tropic,],
                dim=(-2, -1),
            )
        )
        south = torch.sqrt(
            torch.mean(x[..., -ex_tropic:, :] ** 2 * weights[-ex_tropic:], dim=(-2, -1))
        )

        # interpolate in tropics
        tropic = torch.zeros(north.shape[0], 1, 1, nvar, in_tropic).to(device)
        for batch in range(north.shape[0]):
            for var in range(nvar):
                tropic[batch, 0, 0, var] = torch.linspace(
                    north[batch, 0, 0, var], south[batch, 0, 0, var], in_tropic
                )

        # expand and concatenate
        north = north[..., None].repeat(1, 1, 1, 1, ex_tropic)
        south = south[..., None].repeat(1, 1, 1, 1, ex_tropic)

        return torch.cat([north, tropic, south], dim=-1)[..., None]

    def force_non_neg(self, x: torch.Tensor) -> torch.Tensor:
        """Clip variables to non-negative values

        Parameters
        ----------
        x : torch.Tensor
            Input tensor
        """
        nn = x[:, :, :, self.clip_idcs, ...]
        nn = torch.where(nn < 0.0, 0.0, nn)
        x[:, :, :, self.clip_idcs, ...] = nn
        return x

    @torch.inference_mode()
    def __call__(
        self,
        x: torch.Tensor,
        coords: CoordSystem,
    ) -> tuple[torch.Tensor, CoordSystem]:
        """Apply perturbation method

        Parameters
        ----------
        x : torch.Tensor
            Input tensor intended to apply perturbation on, not used in this
            perturbation method
        coords : CoordSystem
            Ordered dict representing coordinate system that describes the tensor.
            Must contain coordinates (Any, "time", "lead_time", "variable", "lat",
            "lon"). Time and lead_time must have size 1.

        Returns
        -------
        tuple[torch.Tensor, CoordSystem]:
            Output tensor and respective coordinate system dictionary
        """
        shape = x.shape
        # Check the required dimensions are present
        # We should probably run the perturbation and then check the output noise can
        # match the size of the input tensor
        handshake_dim(coords, required_dim="time", required_index=-5)
        handshake_dim(coords, required_dim="lead_time", required_index=-4)
        handshake_dim(coords, required_dim="variable", required_index=-3)
        handshake_dim(coords, required_dim="lat", required_index=-2)
        handshake_dim(coords, required_dim="lon", required_index=-1)
        handshake_size(coords, required_dim="time", required_size=1)

        if len(shape) != 5 and len(shape) != 6:
            raise ValueError("Input tensor and coords need 5 or 6 dimensions")

        self.noise_amplitude = self.noise_amplitude.to(x.device)

        # Not the cleanest but works
        if len(shape) == 5:
            batch_size = 1
        if len(shape) == 6:
            batch_size = coords[list(coords.keys())[0]].shape[0]

        # This is some pretty annoying logic to deal with storing the centered
        # perturbation for odd batch sizes... could be worse probably can be better
        # TLDR: if odd batch, store the last generated perturbation for next call
        generator_size = 2 * int((batch_size + 1 - len(self._residual)) // 2)
        # Special case where we can skip the generation
        if generator_size == 0:
            noise = [self._residual.pop()]
        else:
            # Create generator if it doesn't exist
            generator = self.create_generator(
                coords["time"], generator_size=generator_size, device=x.device
            )
            # Generate noise, and concat in batch dimension
            noise = self._residual + [next(generator) for _ in range(generator_size)]
            # If we have an extra output (odd batch size) store it for next call
            try:
                self._residual = [noise.pop(batch_size)]
            except IndexError:
                self._residual = []

        # Sanity checkout for this perturbation
        if len(noise) != batch_size:
            raise ValueError(
                "Seems something went wrong in the perturbation, open an issue with your setup"
            )

        x = torch.cat(noise, dim=0)
        if len(shape) == 5:
            x = x.squeeze(0)

        return x, coords



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES.
# SPDX-FileCopyrightText: All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

import numpy as np
import torch

try:
    from torch_harmonics import InverseRealSHT
except ImportError:
    InverseRealSHT = None
from typing_extensions import Self

from earth2studio.utils import handshake_dim
from earth2studio.utils.imports import check_extra_imports
from earth2studio.utils.type import CoordSystem


class Gaussian:
    """Standard Gaussian peturbation

    Parameters
    ----------
    noise_amplitude : float | Tensor, optional
        Noise amplitude, by default 0.05. If a tensor,
        this must be broadcastable with the input data.
    """

    def __init__(self, noise_amplitude: float | torch.Tensor = 0.05):
        self.noise_amplitude = (
            noise_amplitude
            if isinstance(noise_amplitude, torch.Tensor)
            else torch.Tensor([noise_amplitude])
        )

    @torch.inference_mode()
    def __call__(
        self,
        x: torch.Tensor,
        coords: CoordSystem,
    ) -> tuple[torch.Tensor, CoordSystem]:
        """Apply perturbation method

        Parameters
        ----------
        x : torch.Tensor
            Input tensor intended to apply perturbation on
        coords : CoordSystem
            Ordered dict representing coordinate system that describes the tensor

        Returns
        -------
        tuple[torch.Tensor, CoordSystem]:
            Output tensor and respective coordinate system dictionary
        """
        noise_amplitude = self.noise_amplitude.to(x.device)
        return x + noise_amplitude * torch.randn_like(x), coords


@check_extra_imports("perturbation", [InverseRealSHT])
class CorrelatedSphericalGaussian:
    """Produces Gaussian random field on the sphere with Matern
    covariance peturbation method output to a lat lon grid

    Warning
    -------
    Presently this method generates noise on equirectangular grid of size [N, 2*N] when
    N is even or [N+1, 2*N] when N is odd.

    Parameters
    ----------
    noise_amplitude : float | Tensor, optional
        Noise amplitude, by default 0.05. If a tensor,
        this must be broadcastable with the input data.
    alpha : float, optional
        Regularity parameter. Larger means smoother, by default 2.0
    tau : float, optional
        Length-scale parameter. Larger means more scales, by default 3.0
    sigma : Union[float, None], optional
        Scale parameter. If None, sigma = tau**(0.5*(2*alpha - 2.0)), by default None
    """

    def __init__(
        self,
        noise_amplitude: float | torch.Tensor | None = None,
        sigma: float = 1.0,
        length_scale: float = 5.0e5,
        time_scale: float = 48.0,
    ) -> None:
        if noise_amplitude is None:
            raise ValueError("pass noise amplitude")
        self.sigma = sigma
        self.length_scale = length_scale
        self.time_scale = time_scale
        self.noise_amplitude = (
            noise_amplitude
            if isinstance(noise_amplitude, torch.Tensor)
            else torch.Tensor([noise_amplitude])
        )

    @torch.inference_mode()
    def __call__(
        self,
        x: torch.Tensor,
        coords: CoordSystem,
    ) -> tuple[torch.Tensor, CoordSystem]:
        """Apply perturbation method

        Parameters
        ----------
        x : torch.Tensor
            Input tensor intended to apply perturbation on
        coords : CoordSystem
            Ordered dict representing coordinate system that describes the tensor, must
            contain "lat" and "lon" coordinates

        Returns
        -------
        tuple[torch.Tensor, CoordSystem]:
            Output tensor and respective coordinate system dictionary
        """
        shape = x.shape
        # Check the required dimensions are present
        handshake_dim(coords, required_dim="lat", required_index=-2)
        handshake_dim(coords, required_dim="lon", required_index=-1)

        # Check the ratio
        if 2 * (shape[-2] // 2) != shape[-1] / 2:
            raise ValueError("Lat/lon aspect ration must be N:2N or N+1:2N")

        nlat = 2 * (shape[-2] // 2)  # Noise only support even lat count
        sampler = CorrelatedSphericalField(
            nlat=nlat,
            length_scale=self.length_scale,
            time_scale=self.time_scale,
            sigma=self.sigma,
            N=np.prod(shape[1:-2], dtype=int),
        )
        sampler = sampler.to(x.device)

        sample_noise = sampler(x, None)
        sample_noise = sample_noise.reshape(*shape[:-2], nlat, 2 * nlat)

        # Hack for odd lat coords
        if x.shape[-2] % 2 == 1:
            noise = torch.zeros_like(x)
            noise[..., :-1, :] = sample_noise
            noise[..., -1:, :] = noise[..., -2:-1, :]
        else:
            noise = sample_noise

        noise_amplitude = self.noise_amplitude.to(x.device)
        return x + noise_amplitude * noise, coords


class CorrelatedSphericalField(torch.nn.Module):
    """
    This class was taken from https://github.com/ankurmahesh/earth2mip-fork/blob/HENS/earth2mip/ensemble_utils.py#L392-L531.
    Reference publication: A.Mahesh et al. Huge Ensembles Part I: Design of Ensemble Weather Forecasts using Spherical Fourier Neural Operators https://arxiv.org/abs/2408.03100.

    This class can be used to create noise on the sphere
    with a given length scale (in km) and time scale (in hours).

    It mimics the implementation of the SPPT: Stochastic Perturbed
    Parameterized Tendency in this paper:

    https://www.ecmwf.int/sites/default/files/elibrary/2009/11577-stochastic-parametrization-and-model-uncertainty.pdf

    Parameters
    ----------
    nlat : int
        Number of latitudinal modes;
        longitudinal modes are 2*nlat.
    length_scale : int
        Length scale in km
    time_scale : int
        Time scale for the AR(1) process, that governs
        the evolution of the coefficients
    sigma: float
        desired standard deviation of the field in grid point space
    N: int
        Number of latent dimensions
    grid : string, default is "equiangular"
        Grid type. Currently supports "equiangular" and
        "legendre-gauss".
    dtype : torch.dtype, default is torch.float32
        Numerical type for the calculations.
    """

    def __init__(
        self,
        nlat: int,
        length_scale: float,
        time_scale: float,
        sigma: float,
        N: int,
        grid: str = "equiangular",
        dtype: torch.dtype = torch.float32,
    ) -> None:
        super().__init__()
        self.sigma = sigma
        dt = 6.0
        self.phi = np.exp(-dt / time_scale)

        # Number of latitudinal modes.
        self.nlat = nlat

        # Inverse SHT
        self.isht = InverseRealSHT(
            self.nlat, 2 * self.nlat, grid=grid, norm="backward"
        ).to(dtype=dtype)

        r_earth = 6.371e6
        # kT is defined on slide 7
        self.kT = (length_scale / r_earth) ** 2 / 2
        F0 = self.calculateF0(self.sigma, self.phi, self.nlat, self.kT)

        prods = (
            torch.tensor([j * (j + 1) for j in range(0, self.nlat)])
            .view(self.nlat, 1)
            .repeat(1, self.nlat + 1)
        )

        sigma_n = torch.tril(torch.exp(-self.kT * prods / 2) * F0)
        self.register_buffer("sigma_n", sigma_n)

        # Save mean and var of the standard Gaussian.
        # Need these to re-initialize distribution on a new device.
        mean = torch.tensor([0.0]).to(dtype=dtype)
        var = torch.tensor([1.0]).to(dtype=dtype)
        self.register_buffer("mean", mean)
        self.register_buffer("var", var)
        self.N = N

        # Standard normal noise sampler.
        # Why is this defined in the init? Not sure, would be a lot more flexible
        # in the call function
        self.gaussian_noise = torch.distributions.normal.Normal(self.mean, self.var)
        xi = self.gaussian_noise.sample(
            torch.Size((self.N, self.nlat, self.nlat + 1, 2))
        ).squeeze()
        xi = torch.view_as_complex(xi)

        # Set specrtral cofficients to this value at initial time
        # for stability in teh AR(1) process.  See link in description
        coeff: torch.tensor = ((1 - self.phi**2) ** (-0.5)) * self.sigma_n * xi
        coeff = coeff.unsqueeze(0)
        self.register_buffer("coeff", coeff)

    def calculateF0(
        self, sigma: float, phi: float, nlat: int, kT: float
    ) -> torch.Tensor:
        """
        This function scales the coefficients such that their
        grid-point standard deviation is sigma.
        sigma is the desired variance
        phi is a np.exp(-dt/time_scale)
        """
        numerator = sigma**2 * (1 - (phi**2))
        wavenumbers = torch.arange(1, nlat)
        denominator = (2 * wavenumbers + 1) * torch.exp(
            -kT * wavenumbers * (wavenumbers + 1)
        )
        denominator = 2 * torch.Tensor(denominator).sum()

        return (numerator / denominator) ** 0.5

    def forward(self, x: torch.Tensor, time: np.datetime64 = None) -> torch.Tensor:
        """
        Generate and return a field with a correlated length scale.

        Update the coefficients using an AR(1) process.
        """
        noises = []
        # iterate over samples in batch
        for _ in range(x.shape[0]):
            noise = self.isht(self.coeff) * 4 * np.pi  # type: ignore
            noises.append(noise.reshape(1, 1, 1, self.N, self.nlat, self.nlat * 2))

            # Sample Gaussian noise. # TODO why??? for next step maybe?
            xi = self.gaussian_noise.sample(
                torch.Size((self.N, self.nlat, self.nlat + 1, 2))
            ).squeeze()
            xi = torch.view_as_complex(xi)

            self.coeff = (self.phi * self.coeff) + (self.sigma_n * xi)  # type: ignore

        return torch.cat(noises)

    def to(self, *args: Any, **kwargs: Any) -> Self:
        """Override cuda and to methods so sampler gets initialized with mean and
        variance on the correct device, to(*args, **kwargs)
        """
        super().to(*args, **kwargs)
        self.gaussian_noise = torch.distributions.normal.Normal(self.mean, self.var)

        return self
